// **********************************************************************
//
// Copyright (c) 2001 MT Tools.
//
// All Rights Reserved
//
// MT_DORB is based in part on the product DORB,
// written by Shadrin Victor
//
// See Readme.txt for contact information
//
// **********************************************************************
unit codec;

interface

uses
  orb_int,orb,Classes,types,windows,sysutils,giop_int,addr_int,
  code_int,comp_int,iiop_int,disp_int,req_int;


type

  TGIOPCodec = class(TORBObject,IGIOPCodec)
  private
    FDecoder : IDecoder;
    FEncoder : IEncoder;
    FVersion : _ushort;
    FSize_offset : _ulong;
    FHeader_len : _ulong;
  protected
    function get_encoder(): IEncoder;
    function get_decoder(): IDecoder;
    function put_header(const outc : IGIOPOutContext;mt: TMessageType): _ulong;
    function put_invoke_request(const outc : IGIOPOutContext;mid: MessageID;
      resp:Boolean;const obj: IORBObject;const req: IORBRequest; pr: IPrincipal): Boolean;
    procedure put_contextlist(const outc : IGIOPOutContext;const codesets: Boolean);
    function get_contextlist(con: IGIOPInContext): Boolean;
    procedure put_target(const outc : IGIOPOutContext;const obj : IORBObject);
    procedure put_size(const outc : IGIOPOutContext;start: _ulong);
    function put_args(const outc : IGIOPOutContext;const req: IORBRequest;input: Boolean): Boolean;
    function get_invoke_reply1(const incon: IGIOPInContext; var reqid: _ulong;var status: TReplyStatusType): Boolean;
    function get_invoke_reply2(const incon: IGIOPInContext;reqid: _ulong;
      status: TReplyStatusType;var obj: IORBObject;req: IORBRequest): Boolean;
    function check_header(ctx: IGIOPInContext; var mt: TMessageType; var size: _ulong; var flags: byte ): Boolean;
    function get_header(ctx: IGIOPInContext; var mt: TMessageType; var size: _ulong; var flags: byte ): Boolean;
    function header_len: _ulong;
    function put_bind_request(const outc: IGIOPOutContext;reqid: _ulong; const repoid: string;
      const tag: IObjectTag): Boolean;
    function get_bind_reply(const ctx: IGIOPInContext; out status: TLocateStatus; out obj: IORBObject): Boolean;
    function get_invoke_request(const cin: IGIOPInContext; out reqid: _ulong; out flags: byte; const obj: IORBObject;
      out req: IORBRequest; const pr: IPrincipal): Boolean;
    function get_target(const inc: IGIOPInContext; const obj : IORBObject): Boolean;
    function put_invoke_reply(const outc: IGIOPOutContext; req_id: _ulong; stat: TReplyStatusType;
      obj: IORBObject; req: IORBRequest): Boolean;
    function version(): _ulong;
    function put_invoke_reply_offset(const outc: IGIOPOutContext;const req: IORBRequest): Boolean;
    function put_close_msg(const ctx: IGIOPOutContext): Boolean;
    function put_cancel_request(const ctx: IGIOPOutContext; mid: MessageID): Boolean;
    function get_close_msg(const inc: IGIOPInContext): Boolean;
    function put_error_msg(const ctx: IGIOPOutContext): Boolean;
    function get_bind_request(const ctx: IGIOPInContext; out repoid: string; out tag: IObjectTag): Boolean;
    function put_bind_reply(const outc: IGIOPOutContext; reqid: _ulong; stat: TLocateStatusType; const obj: IORBObject): Boolean;
  public
    constructor Create(dec: IDecoder;enc: IEncoder);
  end;

implementation

uses
  iop,iior_int,codeset_ref,ior,transp,intercept,env_int,env;

{************************************************************
  TGIOPCodec
************************************************************}
constructor TGIOPCodec.Create(dec: IDecoder;enc: IEncoder);
var
  outc : IGIOPOutContext;
begin
  FVersion := $0100;
  FDecoder := dec;
  Fencoder := enc;
  outc := TGIOPOutContext.Create(self,nil,nil);
  fsize_offset := put_header(outc,Request);
  FHeader_len := outc.encoder.buffer.WPosition;
end;

function TGIOPCodec.get_bind_request(const ctx: IGIOPInContext; out repoid: string; out tag: IObjectTag): Boolean;
var
  dec: IDecoder;
  l: _ulong;
  str: string;
begin
  result := false;
  dec := ctx.decoder;
  if not dec.get_string(repoid) then exit;
  if not dec.seq_begin(l) then exit;
  if l > 0 then
    begin
      SetLength(str,l);
      if not dec.get_octets(PChar(str)^,l) then exit;
      tag := TObjectTag.Create(str);
    end
  else
    tag := TObjectTag.Create('');
  if not dec.seq_end then exit;
  result := true;
end;

function TGIOPCodec.put_error_msg(const ctx: IGIOPOutContext): Boolean;
var
  enc: IEncoder;
  l : _ulong;
begin
  enc := ctx.encoder;
  l := put_header(ctx,MessageError);
  put_size(ctx,l);
  result := true;
end;

function TGIOPCodec.get_close_msg(const inc: IGIOPInContext): Boolean;
begin
  FDecoder := inc.decoder;
  result := true;
end;

function TGIOPCodec.get_header(ctx: IGIOPInContext; var mt: TMessageType; var size: _ulong; var flags: byte ): Boolean;
var
  dec: IDecoder;
  magic : array[0..4] of char;
  ver : _ushort;
  min,maj,order : byte;

begin
  result := false;
  dec := ctx.decoder;
  magic[4] := Char(0);
  if dec.get_octets(magic,4) and (StrComp(@magic,'GIOP')= 0) then
    begin
      if not dec.get_octet(Octet(maj)) then exit;
      if not dec.get_octet(Octet(min)) then exit;
      ver := ((maj shl 8) or min);
      if ver <= $1020 then
        begin
          FVersion := ver;
          if FVersion = $0100 then
            begin
              if not dec.get_octet(Octet(order)) then exit;
              if order <> 0 then
                dec.byteorder(types.LittleEndian)
              else
                dec.byteorder(types.BigEndian);
              if order <> 0 then flags := 1 else flags := 0;
            end
          else
            begin
              if not dec.get_octet(Octet(flags)) then exit;
              if (flags and GIOP_BYTEORDER_BIT) <> 0 then
                dec.byteorder(types.LittleEndian)
              else
                dec.byteorder(types.BigEndian);
            end;
          if not dec.get_octet(Octet(order)) then exit;
          mt := TMessageType(order);
          result := dec.get_ulong(size);
        end;
    end;
end;

function TGIOPCodec.check_header(ctx: IGIOPInContext; var mt: TMessageType; var size: _ulong; var flags: byte ): Boolean;
var
  pos : _ulong;
begin
  pos := ctx.decoder.Buffer.RPosition;
  result := get_header(ctx,mt,size,flags);
  ctx.decoder.Buffer.RseekFromBegin(pos);
end;

function TGIOPCodec.get_decoder(): IDecoder;
begin
  result := FDecoder;
end;

function TGIOPCodec.get_target(const inc: IGIOPInContext; const obj : IORBObject): Boolean;
var
  dec: IDecoder;
  objkey: PByteArray;
  keyval : string;
  keylen: _ulong;
begin
  result := false;
  dec := inc.decoder;
  if FVersion < $0102 then
    begin
      obj.IOR.add_profile(TGIOPSimpleProfile.Create('',0));
      if not dec.seq_begin(keylen) then exit;
      objkey := dec.Buffer.Data;
      if not dec.Buffer.RseekFromCurrent(keylen) then exit;
      if not dec.seq_end then exit;
      SetLength(keyval,keylen);
      move(objkey^,pchar(keyval)^,keylen);
      obj.IOR.set_objectkey(keyval,keylen);
    end
  else
    assert(false,'not implemented');
  result := true;
end;

function TGIOPCodec.get_invoke_request(const cin: IGIOPInContext; out reqid: _ulong; out flags: byte; const obj: IORBObject;
      out req: IORBRequest; const pr: IPrincipal): Boolean;
var
  dec: IDecoder;
  tmp : array[0..2] of byte;
  opname: string;
  env: IEnvironment;
begin
  result := false;
  dec := cin.decoder;
  //ServerInterceptor
  if not TServerInterceptor._idle() then begin
    env := TEnvironment.Create();
    if not TServerInterceptor._exec_input_message(dec.buffer(), env) then begin
      result := false;
      Exit;
    end;
  end;
  if FVersion < $0102 then
    begin
      if not get_contextlist(cin) then exit;
      if not dec.get_ulong(reqid) then exit;
      if not dec.get_octet(Octet(flags)) then exit;
      if (FVersion <> $0100) and not dec.get_octets(tmp,3) then exit;
      if not get_target(cin,obj) then exit;
      if not dec.get_string(opname) then exit;
      if not dec.get_principal(pr) then exit;
    end
  else
    assert(false,'not implemented');
  req := TGIOPRequest.Create(opname,dec,self);
  result := true;
end;

function TGIOPCodec.get_invoke_reply1(const incon: IGIOPInContext; var reqid: _ulong;var status: TReplyStatusType): Boolean;
var
  dec : IDecoder;
  enum : _ulong;
  env: IEnvironment;
begin
  result := false;
  dec := incon.decoder;
  //ClientInterceptor
  if not TClientInterceptor._idle() then begin
    env := TEnvironment.Create();
    if not TClientInterceptor._exec_input_message(dec.buffer(), env) then begin
      result := false;
      Exit;
    end;
  end;
  if FVersion < $0102 then
    begin
      if not get_contextlist(incon) then exit;
      if not dec.get_ulong(reqid) then exit;
      if not dec.get_ulong(enum) then exit;
      status := TReplyStatusType(enum);
    end
  else
    begin
      if not dec.get_ulong(reqid) then exit;
      if not dec.get_ulong(enum) then exit;
      status := TReplyStatusType(enum);
      if not get_contextlist(incon) then exit;
    end;
    {if _giop_ver >= $0102 then  // if there are no arguments aligning might fail ...
      bu->buffer()->ralign (dc->max_alignment());}
  result := true;
end;

function TGIOPCodec.get_invoke_reply2(const incon: IGIOPInContext;reqid: _ulong;
      status: TReplyStatusType;var obj: IORBObject;req: IORBRequest): Boolean;
var
  dec : IDecoder;
begin
  result := false;
  dec := incon.decoder;
  obj := nil;
  case status of
    types.NO_EXCEPTION: if req <> nil then
        result := req.set_out_args(dec,false);
    types.USER_EXCEPTION,types.SYSTEM_EXCEPTION: if req <> nil then
      result := req.set_out_args(dec,true);
    else
      assert(false,'not implemented');
  end;
end;

function TGIOPCodec.get_bind_reply(const ctx: IGIOPInContext; out status: TLocateStatus; out obj: IORBObject): Boolean;
var
  dec: IDecoder;
  ior: IIOR;
  enum: _ulong;
begin
  result := false;
  dec := ctx.decoder;
  obj := nil;
  if not dec.get_ulong(enum) then exit;
  status := TLocateStatus(enum);
  ior := TIOR.Create;
  if not ior.decode(dec) then exit;
  obj := ORB_Instance.ior_to_object(ior);
  result := true;
end;

function TGIOPCodec.header_len: _ulong;
begin
  result := FHeader_len
end;

function TGIOPCodec.put_header(const outc : IGIOPOutContext;mt: TMessageType): _ulong;
var
  enc : IEncoder;
begin
  enc := outc.encoder;
  enc.put_octets('GIOP',4);
  enc.put_octet(Octet(FVersion shr 8));
  enc.put_octet(Octet(FVersion xor $FF00));
  enc.put_octet(Octet(enc.byteorder = LittleEndian));
  enc.put_Octet(Octet(Ord(mt)));
  result := enc.buffer.WPosition;
  enc.put_ulong(0);
end;

function TGIOPCodec.get_encoder(): IEncoder;
begin
  result := FEncoder;
end;

procedure TGIOPCodec.put_size(const outc : IGIOPOutContext;start: _ulong);
var
  enc : IEncoder;
  endpos : _ulong;
begin
  enc := outc.encoder;
  endpos := enc.buffer.WPosition;
  enc.buffer.WseekFromBegin(start);
  enc.put_ulong(endpos - enc.buffer.getRAlignBase - FHeader_len);
  enc.buffer.WseekFromBegin(endpos);
end;

function TGIOPCodec.put_bind_request(const outc: IGIOPOutContext;reqid: _ulong; const repoid: string;
      const tag: IObjectTag): Boolean;
var
  enc : IEncoder;
  key : _ulong;
  obj : IORBObject;
  val : longint;
  env: IEnvironment;
begin
  val := 0;
  enc := outc.encoder;
  key := put_header(outc,Request);
  obj := TORBObject.Create(TIOR.Create);
  obj.IOR.add_profile(TGIOPSimpleProfile.Create());
  if FVersion < $0102 then
    begin
      put_contextlist(outc,false);
      enc.put_ulong(reqid);
      enc.put_octet(Octet(1)); // response flags
      if FVersion <> $0100 then
        enc.put_octets(val,3);
      put_target(outc,obj);
      enc.put_string('_bind');
      enc.put_principal(TPrincipal.Create());
      enc.seq_end;
    end
  else
    begin
      enc.put_ulong(reqid);
      enc.put_octet(Octet(1)); // response flags
      val := 0;
      enc.put_octets(val,3);
      put_target(outc,obj);
      enc.put_string('_bind');
      put_contextlist(outc,false);
    end;
  enc.put_string(repoid);
  enc.seq_begin(tag.get_length);
  if tag.get_length > 0 then
    enc.put_octets(PChar(tag.get_bytes)^,tag.get_length);
  enc.seq_end;
  //ClientInterceptor
  if not TClientInterceptor._idle() then begin
    env := TEnvironment.Create();
    enc.buffer().RseekFromCurrent(header_len());
    if not TClientInterceptor._exec_output_message(enc.buffer(), env) then begin
      result := false;
      Exit;
    end;
    enc.buffer().RseekFromCurrent(-header_len());
  end;
  put_size(outc,key);
  result := true;
end;

function TGIOPCodec.version(): _ulong;
begin
  result := FVersion;
end;

function TGIOPCodec.put_invoke_request(const outc : IGIOPOutContext;mid: MessageID;
      resp: Boolean;const obj: IORBObject;const req: IORBRequest; pr: IPrincipal): Boolean;
var
  enc : IEncoder;
  key : _ulong;
  val : longint;
  //bo : TByteOrder;
  env: IEnvironment;
begin
  val := 0;
  enc := outc.encoder;
  //bo := enc.byteorder;
  if req.req_type = 'giop' then
    enc.byteorder((req as IGIOPRequest).input_byteorder);
  key := put_header(outc,Request);
  if FVersion < $0102 then
    begin
      put_contextlist(outc,true);
      enc.put_ulong(mid);
      enc.put_boolean(resp); // response flags
      if FVersion <> $0100 then
        enc.put_octets(val,3);
      put_target(outc,obj);
      enc.put_string(req.op_name);
      if pr <> nil then
        enc.put_principal(pr)
      else
        enc.put_principal(TPrincipal.Create());
    end
  else
    begin
      enc.put_ulong(mid);
      enc.put_octet(Octet(1)); // response flags
      enc.put_octets(val,3);
      put_target(outc,obj);
      enc.put_string(req.op_name);
      put_contextlist(outc,true);
    end;
  result := put_args(outc,req,true);
  //ClientInterceptor
  if not TClientInterceptor._idle() then begin
    env := TEnvironment.Create();
    enc.buffer().RseekFromCurrent(header_len());
    if not TClientInterceptor._exec_output_message(enc.buffer(), env) then begin
      result := false;
      Exit;
    end;
    enc.buffer().RseekFromCurrent(-header_len());
  end;
  put_size(outc,key);
end;

{**
  Puts argument into request arg`s list
**}
function TGIOPCodec.put_args(const outc : IGIOPOutContext;const req: IORBRequest;input: Boolean): Boolean;
var
  enc : IEncoder;
  is_except: Boolean;
begin
  enc := outc.encoder;
  if input then
    result := req.get_in_args(enc)
  else
    result := req.get_out_args(enc,is_except);
end;

procedure TGIOPCodec.put_target(const outc : IGIOPOutContext;const obj : IORBObject);
var
  enc : IEncoder;
  prof : IIORProfile;
  objkey : string;
  ok_len : _ulong;
begin
  enc := outc.encoder;
  prof := obj.fwd_ior.get_profile();
  objkey := prof.objkey(ok_len);
  if FVersion < $0102 then
    begin
      enc.seq_begin(ok_len);
      enc.put_octets(PChar(objkey)^,ok_len);
      enc.seq_end;
    end
  else
    begin
      enc.put_short(GIOP_KEYADDR);
      enc.seq_begin(ok_len);
      enc.put_octets(PChar(objkey)^,ok_len);
      enc.seq_end;
    end;
end;

{**
  Stores context list in request
**}
procedure TGIOPCodec.put_contextlist(const outc : IGIOPOutContext;const codesets : Boolean);
var
  enc : IEncoder;
  cs,wcs : CodesetID;
  state : TEncapsState;
//  i : integer;
begin
  enc := outc.encoder;
  enc.seq_begin(0+_ulong(codesets));
  //for i := 0 to Pred(0) do
  //  begin
      // must store contextdata
  //  end;
  if codesets then
    begin
      cs := TCodesets[0].id;
      wcs := TCodesets[4].id;
      enc.put_ulong(TAG_CODE_SETS);
      enc.encaps_begin(state);
      enc.put_ulong(cs);
      enc.put_ulong(wcs);
      enc.encaps_end(state);
    end;
  enc.seq_end;
end;

{**
  not implemented
**}
function TGIOPCodec.get_contextlist(con: IGIOPInContext): Boolean;
var
  dec: IDecoder;
  len,size,context_id: _ulong;
  i : integer;
  state: TEncapsState;
  csid,wcsid : CodesetID;
begin
  result := false;
  dec := con.decoder;
  if not dec.seq_begin(len) then exit;
  if len > 0 then
    begin
      for i := 0 to len - 1 do
        begin
          if not dec.get_ulong(context_id) then exit;
          dec.encaps_begin(state,size);
          if context_id = TAG_CODE_SETS then
            begin
              if not dec.get_ulong(csid) then exit;
              if not dec.get_ulong(wcsid) then exit;
            end
          else
            begin
              assert(false,'not implemented');
            end;
          if not dec.encaps_end(state) then exit;
        end;
    end;
  if not dec.seq_end then exit;
  result := true;
end;

function TGIOPCodec.put_bind_reply(const outc: IGIOPOutContext; reqid: _ulong; stat: TLocateStatusType; const obj: IORBObject): Boolean;
var
  enc: IEncoder;
  key: _ulong;
  env: IEnvironment;
begin
  enc := outc.encoder;
  key := put_header(outc,Reply);
  if FVersion < $0102 then
    begin
      put_contextlist(outc,false);
      enc.put_ulong(reqid);
      enc.put_ulong(Ord(NO_EXCEPTION));
    end
  else
    begin
      enc.put_ulong(reqid);
      enc.put_ulong(Ord(NO_EXCEPTION));
      put_contextlist(outc,false);
    end;
  enc.put_ulong(ord(stat));
  if stat = OBJECT_HERE then
    obj.ior.encode(enc)
  else
    IIOR(TIOR.Create()).encode(enc);
  //ServerInterceptor
  if not TServerInterceptor._idle() then begin
    env := TEnvironment.Create();
    enc.buffer().RseekFromCurrent(header_len());
    if not TServerInterceptor._exec_output_message(enc.buffer(), env) then begin
      result := false;
      Exit;
    end;
    enc.buffer().RseekFromCurrent(-header_len());
  end;
  put_size(outc,key);
  result := true;
end;

function TGIOPCodec.put_invoke_reply(const outc: IGIOPOutContext; req_id: _ulong; stat: TReplyStatusType;
      obj: IORBObject; req: IORBRequest): Boolean;
var
  enc: IEncoder;
  key : _ulong;
  env: IEnvironment;
begin
  result := false;
  enc := outc.encoder;
  key := put_header(outc,types.Reply);
  if FVersion < $0102 then
    begin
      put_contextlist(outc,false);
      enc.put_ulong(req_id);
      enc.put_ulong(Ord(stat));
    end
  else
    begin
      enc.put_ulong(req_id);
      enc.put_ulong(Ord(stat));
      put_contextlist(outc,false);
    end;
  case stat of
    NO_EXCEPTION,USER_EXCEPTION,SYSTEM_EXCEPTION: begin
        if FVersion > $0102 then
          begin
            assert(false,'not implemented');
          end;
        if not put_args(outc,req,false) then
          exit;
      end;
    else
      assert(false,'not implemented');
  end;
  //ServerInterceptor
  if not TServerInterceptor._idle() then begin
    env := TEnvironment.Create();
    enc.buffer().RseekFromCurrent(header_len());
    if not TServerInterceptor._exec_output_message(enc.buffer(), env) then begin
      result := false;
      Exit;
    end;
    enc.buffer().RseekFromCurrent(-header_len());
  end;
  put_size(outc,key);
  result := true;
end;

function TGIOPCodec.put_invoke_reply_offset(const outc: IGIOPOutContext;const req: IORBRequest): Boolean;
var
  enc: IEncoder;
begin
  result := false;
  enc := outc.encoder;
  enc.buffer.WseekFromCurrent(FHeader_len); //?????
  put_contextlist(outc,false);
  enc.put_ulong(0);
  enc.put_ulong(0);
end;

function TGIOPCodec.put_close_msg(const ctx: IGIOPOutContext): Boolean;
var
  l : _ulong;
begin
  l := put_header(ctx,CloseConnection);
  put_size(ctx,l);
  result := true;
end;

function TGIOPCodec.put_cancel_request(const ctx: IGIOPOutContext; mid: MessageID): Boolean;
var
  enc: IEncoder;
  l : _ulong;
begin
  enc := ctx.encoder;
  l := put_header(ctx,CancelRequest);
  enc.put_ulong(mid);
  put_size (ctx,l);
  result := true;
end;


end.
