// **********************************************************************
//
// Copyright (c) 2001 MT Tools.
//
// All Rights Reserved
//
// MT_DORB is based in part on the product DORB,
// written by Shadrin Victor
//
// See Readme.txt for contact information
//
// **********************************************************************
{**
   Implementation of IIOPServerInvokeRecord and IIOPServer.
 }
unit iiop;

interface

uses
  iiop_int,orb_int,types,orb,giop_int,addr_int,classes,req_int,map_int,
  env_int,code_int,OSThread,mtmanager_int,iop;

{$I ..\dorb.inc}

type

  TIIOPServerInvokeRecord = class(TInterfacedObject,IIIOPServerInvokeRecord)
  private
    FConnection: IGIOPConnection;
    FORBId : ORBMessageID;
    FREQId : MessageID;
    FRequest : IORBRequest;
    FObject: IORBObject;
    FActive: boolean;
    FOrbMsgId: MessageID;
  protected
    procedure init_invoke (const conn: IGIOPConnection; reqid: MessageID;
      orbid: ORBMessageID; const req : IORBRequest; const obj: IORBObject);
    function get_request(): IORBRequest;
    function orbmsgid(): MessageID;
    function orbid: ORBMessageID;
    function reqid: MessageID;
    procedure set_orbid(oid: ORBMessageID);
    procedure set_reqid(reqid: MessageID);
    function connection : IGIOPConnection;
    procedure clean();
    function active(): boolean;
    procedure deactivate();
  public
  end;

  TIIOPServer = class(TORBObject,IIIOPServer,IObjectAdaptor,ITransportServerCallback,IGIOPCallBack,IORBCallback)
  private
    FTServers : IInterfaceList;
    FORB : IORB;
{$IFDEF USE_IOP_CACHE}
    FCacheRecord : IIIOPServerInvokeRecord;
    FCacheUsed: Boolean;
{$ENDIF}
    FConnections : IInterfaceList;
    FORBIds : ILongInterfaceMap;
    FORBIdsLock: TMutex;
    FConnLock: TMutex;
    FConnMgr: TGIOPConnectionMgr;
    procedure deref_connection(const conn: IGIOPConnection; all: boolean = false);
  protected
    function skeleton(const obj: IORBObject): IORBObject;
    function is_local(): Boolean;
    procedure add_invoke(rec: IIIOPServerInvokeRecord);
    procedure del_invoke_orbid(rec : IIIOPServerInvokeRecord);
    function pull_invoke_orbid(orbid: ORBMessageID):IIIOPServerInvokeRecord;
    procedure del_invoke_reqid(orbid: MessageID;const conn: IGIOPConnection);
    procedure kill_connection(const conn: IGIOPConnection; redo: boolean = false);
    procedure close_connection(const conn: IGIOPConnection);
    function handle_input(const conn: IGIOPConnection; const inp: IBuffer): Boolean;
    function exec_invoke_request(const inc: IGIOPInContext;obj : IORBObject;
      req: IORBRequest;  const pr: IPrincipal; resp_exp: Boolean;const conn: IGIOPConnection; id: ORBMessageID): ORBMessageID;
    function handle_invoke_request (const conn: IGIOPConnection;const inc: IGIOPInContext):Boolean;
    function handle_locate_request (const conn: IGIOPConnection;const inc: IGIOPInContext):Boolean;
    function handle_cancel_request (const conn: IGIOPConnection;const inc: IGIOPInContext):Boolean;
    procedure handle_invoke_reply (id: ORBMessageID);
    procedure handle_locate_reply (msgid: MessageID);
    procedure handle_bind_reply (id: ORBMessageID);
    function listen(): Boolean;
    function listen_addr(addr: IAddress;addr2: IAddress):Boolean;
    function has_object(const obj: IORBObject):Boolean;
    function waitfor(const orb: IORB; id: ORBMessageID; event: TORBCallbackEvent; tmout: long = -1): boolean;
    procedure notify(const orb: IORB; id: ORBMessageID; evt : TORBCallbackEvent);
    procedure ts_callback(ts: ITransportServer;event :TTransportServerCallBackEvent);
    //procedure callback (const orb: IORB; mid: MessageID; evt : TORBCallbackEvent);
    function invoke(mid: ORBMessageID;const obj : IORBObject; const req : IORBRequest; const pr: IPrincipal; response_exp: Boolean): Boolean; stdcall;
    function bind(mid: ORBMessageID; const repoid: string; const oid: IObjectTag; const addr: IAddress): Boolean;
    function get_oaid(): ObjectID;
    function giop_callback(const conn: IGIOPConnection;event: TEventType): Boolean;
    function input_giop_callback(const conn: IGIOPConnection;const buf: IBuffer{;event: TEventType}): Boolean;
    function create_invoke(): IIIOPServerInvokeRecord;
    procedure answer_invoke(mid: ORBMessageID; const obj: IORBObject; const req: IORBRequest; stat: TInvokeStatus);
    procedure shutdown(WaitForCompletion: Boolean);
    procedure conn_closed(const conn: IGIOPConnection);
    procedure cancel(mid: ORBMessageID);
    procedure abort_invoke_orbid(rec : IIIOPServerInvokeRecord);
    procedure sendORBMsg(const conn: IGIOPConnection;event: TORBMsgEvent);
  public
    constructor Create(dorb: IORB);
    destructor Destroy(); override;
  end;

implementation

uses
  address,iior_int,orbcomp,mcomp_int,codec,decoder,encoder,
  ior,map,utils, intercept,mtmanager;

//**************************************************************************
//  TIIOPServer
//**************************************************************************
constructor TIIOPServer.Create(dorb: IORB);
begin
  inherited Create();
  FORBIdsLock := TMutex.Create(false, maRecursive);
  FConnLock := TMutex.Create;
  FTServers := TInterfaceList.Create;
  FConnections := TInterfaceList.Create;
  FORBIds := TLongInterfaceMap.Create;
  FConnMgr := TGIOPConnectionMgr.Create;
{$IFDEF USE_IOP_CACHE}
  FCacheUsed := false;
  FCacheRecord := TIIOPServerInvokeRecord.Create;
{$ENDIF}
  { TODO -oOVS : this has to accessible from autoside }
  FConnMgr.set_conn_limit(10);
  FORB := dorb;
  FORB.register_oa(self,OAPriorityRemote);
end;

destructor TIIOPServer.Destroy();
begin
  FConnLock.lock();
  try
    FConnections := nil;
  finally
    FConnLock.unlock();
  end; { try/finally }

  FTServers := nil;
  
  FConnLock.lock();
  try
    FORBIDs := nil;
  finally
    FConnLock.unlock();
  end; { try/finally }
  FORBIdsLock.Free;
  FConnLock.Free;
  FConnMgr.Free;
  inherited Destroy();
end;

procedure TIIOPServer.conn_closed(const conn: IGIOPConnection);
var
  outc : IGIOPOutContext;
begin
  outc := TGIOPOutContext.Create(conn.codec,nil,nil);
  conn.codec.put_close_msg(outc);
  deref_connection(conn, true);
end;

procedure TIIOPServer.shutdown(WaitForCompletion: Boolean);
var
  i : integer;
begin
  {for i := 0 to Pred(FTServers.Count) do
    ITransportServer(FTServers[i]).aselect(FORB.dispatcher,nil);}
  FTServers.Clear;
  FConnLock.lock();
  try
    for i := 0 to Pred(FConnections.Count) do
      conn_closed(IGIOPConnection(FConnections[i]));
    FConnections.Clear;
  finally
    FConnLock.unlock();
  end; { try/finally }
{$IFDEF USE_IOP_CACHE}
  if FCacheUsed then
    FORB.cancel(FCacheRecord.orbid);
{$ENDIF}
  FORBIdsLock.lock();
  try
    for i := Pred(FORBIDs.Count) downto 0 do
      FORB.cancel(IIIOPServerInvokeRecord(FORBIDs[i]).orbid);
    FORBIDs.Clear;
  finally
    FORBIdsLock.unlock();
  end; { try/finally }
{$IFDEF USE_IOP_CACHE}
  FCacheUsed := false;
{$ENDIF}
  FORB.answer_shutdown(self);
end;

procedure TIIOPServer.answer_invoke(mid: ORBMessageID; const obj: IORBObject; const req: IORBRequest; stat: TInvokeStatus);
begin
  assert(false,'not implemented');
end;

function TIIOPServer.skeleton(const obj: IORBObject): IORBObject;
begin
  result := nil;
end;

procedure TIIOPServer.deref_connection(const conn: IGIOPConnection;
  all: boolean);
begin
  if conn.deref() then begin
    sendORBMsg(conn, omeKillConn);
    FConnMgr.release_conn();
  end;
end;

procedure TIIOPServer.abort_invoke_orbid(rec : IIIOPServerInvokeRecord);
begin
  FORB.cancel(rec.orbmsgid());
  //del_invoke_orbid(mid);
end;

procedure TIIOPServer.sendORBMsg(const conn: IGIOPConnection;
  event: TORBMsgEvent);
var
  msg: IMsgType;
  orbmsg: IOrbMsg;
begin
  orbmsg := TORBMsg.Create(conn, nil, event);
  orbmsg._AddRef();
  msg := TMsgType.Create(Pointer(orbmsg));
  TPM.put_msg(ord(odORB), msg);
end;

function TIIOPServer.is_local(): Boolean;
begin
  result := false;
end;

function TIIOPServer.get_oaid(): ObjectID;
begin
  result := 'dorb-iiop-server';
end;

function TIIOPServer.giop_callback(const conn: IGIOPConnection;event: TEventType): Boolean;
var
  addr : IAddress;
begin
  result := false;
  case event of
    types.InputReady: result := handle_input(conn, conn.input);
    types.Idle: begin
        close_connection(conn);
        kill_connection(conn);
      end;
    types.Closed: begin
        //conninterceptor
        addr := conn.transport().peer();
        Assert(addr <> nil);
        TConnInterceptor._exec_client_disconnect(addr.toString);
        kill_connection(conn);
      end;
    else
      assert(false,'not implemented');
  end;
end;

function TIIOPServer.input_giop_callback(const conn: IGIOPConnection;
  const buf: IBuffer{; event: TEventType}): Boolean;
begin
  result := handle_input(conn, buf);
end;

procedure TIIOPServer.ts_callback(ts: ITransportServer;event :TTransportServerCallBackEvent);
var
  transp : ITransport;
  addr : IAddress;
  conn : IGIOPConnection;
  res: boolean;
begin
  case event of
    types.tsAccept: begin
        transp := ts.accept;
        if transp <> nil then begin
          if not FConnMgr.aquire_conn() then exit;
          if transp.bad then  exit;
          addr := transp.peer;
          //conninterceptor
          Assert(addr <> nil);
          res := TConnInterceptor._exec_client_connect(addr.toString);
          if res then begin
            conn := TGIOPConnection.Create(FConnMgr.Dispatcher(),FORB,transp,self,TGIOPCodec.Create(TDecoder.Create(nil),TEncoder.Create(nil)));
            FConnLock.lock();
            try
              FConnections.Add(conn);
            finally
              FConnLock.unlock();
            end; { try/finally }
            conn.start();
          end;
        end;
      end;
    else
      assert(false,'not implemented');
  end;
end;

function TIIOPServer.create_invoke(): IIIOPServerInvokeRecord;
begin
{$IFDEF USE_IOP_CACHE}
  if FCacheUsed then
    result := TIIOPServerInvokeRecord.Create
  else
    begin
      FCacheUsed := True;
      result := FCacheRecord;
    end;
{$ELSE}
  result := TIIOPServerInvokeRecord.Create;
{$ENDIF}
end;

procedure TIIOPServer.add_invoke(rec: IIIOPServerInvokeRecord);
var
  al: IAuto_Lock;
begin
  al := TAuto_Lock.Create(FORBIdsLock);
{$IFDEF USE_IOP_CACHE}
  if rec <> FCacheRecord then
    FORBIDs[rec.orbmsgid] := rec;
{$ELSE}
  FORBIDs[rec.orbmsgid] := rec;
  rec._AddRef;
  FORB.set_RequestHint(rec.orbid(), Pointer(rec));
{$ENDIF}
end;

procedure TIIOPServer.notify(const orb: IORB; id: ORBMessageID;
  evt: TORBCallbackEvent);
begin
  case evt of
    _Invoke: handle_invoke_reply(id);
    _Bind: handle_bind_reply(id);
    else
      assert(false,'not implemented');
  end;
end;

function TIIOPServer.waitfor(const orb: IORB; id: ORBMessageID;
  event: TORBCallbackEvent; tmout: Integer): boolean;
begin
  Assert(false, 'not implemented');
  result := true;
end;

procedure TIIOPServer.close_connection(const conn: IGIOPConnection);
begin
  assert(false,'not implemented');
end;

function TIIOPServer.pull_invoke_orbid(orbid: ORBMessageID):IIIOPServerInvokeRecord;
var
  al: IAuto_Lock;
begin
  result := nil;
{$IFDEF USE_IOP_CACHE}
  if FCacheUsed and (FCacheRecord.orbid = orbid) then
    result := FCacheRecord
  else
    result := FORBIDs[orbid] as IIIOPServerInvokeRecord;
{$ENDIF}
  al := TAuto_Lock.Create(FORBIdsLock);
  result := IIIOPServerInvokeRecord(FORB.get_RequestHint(orbid));
  result._Release;
  if (result <> nil) and result.active() then begin
    result.deactivate();
    Exit;
  end;
  del_invoke_orbid(result);
end;

procedure TIIOPServer.del_invoke_orbid(rec : IIIOPServerInvokeRecord);
var
  al: IAuto_Lock;
begin
  al := TAuto_Lock.Create(FORBIdsLock);
  Assert(not rec.active());
{$IFDEF USE_IOP_CACHE}
  if FCacheUsed and (FCacheRecord.orbid = orbid) then
    begin
      FCacheRecord.clean;
      FCacheUsed := false;
      exit;
    end;
{$ELSE}
  deref_connection(rec.connection());
  if rec <> nil then
    begin
      rec.clean;
      FORBIDs.erase(rec.orbmsgid());
    end;
{$ENDIF}
end;

procedure TIIOPServer.del_invoke_reqid(orbid: MessageID;
  const conn: IGIOPConnection);
begin
  assert(false,'not implemented');
end;

function TIIOPServer.exec_invoke_request(const inc: IGIOPInContext;
  obj: IORBObject; req: IORBRequest;  const pr: IPrincipal; resp_exp: Boolean;
  const conn: IGIOPConnection; id: ORBMessageID): ORBMessageID;
var
  rep : string;
  tag : IObjectTag;
begin
  if req.op_name = '_bind' then
    begin
      conn.codec.get_bind_request(inc,rep,tag);
      result := Forb.bind_async(PChar(rep),tag,nil,self,id);
    end
  else
    result := Forb.invoke_async(obj,req,pr,resp_exp,self,id);
end;

procedure TIIOPServer.handle_bind_reply(id: ORBMessageID);
var
  obj : IORBObject;
  stat: TLocateStatus;
  rec : IIIOPServerInvokeRecord;
  giop_stat : TLocateStatusType;
  outc: IGIOPOutContext;
begin
  rec := pull_invoke_orbid(id);
  stat := forb.get_bind_reply(id,obj);
  if rec = nil then exit;
  giop_stat := OBJECT_HERE;
  case stat of
    LocateUnknown: giop_stat := UNKNOWN_OBJECT;
    LocateHere : giop_stat := OBJECT_HERE;
    LocateForward : giop_stat := OBJECT_FORWARD;
  end;
  outc := TGIOPOutContext.Create(rec.connection.codec);
  rec.connection.codec.put_bind_reply(outc,rec.reqid,giop_stat,obj);
  rec.connection.output(outc.buffer);
  del_invoke_orbid(rec);
end;

function TIIOPServer.handle_cancel_request(const conn: IGIOPConnection;
  const inc: IGIOPInContext): Boolean;
begin
  result := false;
  assert(false,'not implemented');
end;

function TIIOPServer.handle_input(const conn: IGIOPConnection; const inp: IBuffer): Boolean;
var
  inc : IGIOPInContext;
  size: _ulong;
  flags: byte;
  mt: TMessageType;
begin
  result := false;
  inc := TGIOPInContext.Create(conn.codec,inp,nil,nil);
  if not conn.codec.get_header(inc,mt,size,flags) then exit;
  case mt of
    types.Request: result := handle_invoke_request(conn,inc);
    else begin
      conn.activeDeref();
      assert(false,'not implemented');
    end;
  end;
end;

procedure TIIOPServer.handle_invoke_reply(id: ORBMessageID);
var
  obj : IORBObject;
  stat : TInvokeStatus;
  req: IORBRequest;
  rec: IIIOPServerInvokeRecord;
  gstat : TReplyStatusType;
  outc: IGIOPOutContext;
begin
  stat := FOrb.get_invoke_reply(id,obj,req);
  rec := pull_invoke_orbid(id);
  if rec = nil then exit;
  gstat := types.NO_EXCEPTION;
  case stat of
    InvokeOk : gstat := types.NO_EXCEPTION;
    InvokeSysEx: gstat := types.SYSTEM_EXCEPTION;
    InvokeUserEx: gstat := types.USER_EXCEPTION;
    InvokeForward: gstat := types.LOCATION_FORWARD;
  end;
  outc := TGIOPOutContext.Create(rec.connection.codec,nil,nil);
  rec.connection.codec.put_invoke_reply(outc,rec.reqid,gstat,obj,req);
  rec.connection.output(outc.buffer);
  del_invoke_orbid(rec);
end;

function TIIOPServer.handle_invoke_request(const conn: IGIOPConnection;
  const inc: IGIOPInContext): Boolean;
var
  obj : IORBObject;
  reqid: _ulong;
  resp_exp : Byte;
  req: IORBRequest;
  msgid: MessageID;
  orbid: ORBMessageID;
  rec : IIIOPServerInvokeRecord;
  pr : IPrincipal;
begin
  result := false;
  obj := TORBObject.Create(TIOR.Create);
  pr := conn.transport().get_principal;
  if not conn.codec.get_invoke_request(inc,reqid,resp_exp,obj,req,pr) then begin
    conn.activeDeref();
    conn.deref();
    exit;
  end;
  msgid := FORB.new_messageid;
  if Boolean(resp_exp) then
    begin
      orbid := FORB.new_orbid(msgid);
      conn.ref();
      rec := create_invoke();
      rec.init_invoke(conn,reqid,orbid,req,obj);
      add_invoke(rec);
    end;
  conn.activeDeref();
  exec_invoke_request(inc,obj,req,pr,Boolean(resp_exp),conn,orbid);
end;

procedure TIIOPServer.handle_locate_reply(msgid: MessageID);
begin
  assert(false,'not implemented');
end;

function TIIOPServer.handle_locate_request(const conn: IGIOPConnection;
  const inc: IGIOPInContext): Boolean;
begin
  result := false;
  assert(false,'not implemented');
end;

function TIIOPServer.has_object(const obj: IORBObject): Boolean;
begin
  result := false;
end;

procedure TIIOPServer.kill_connection(const conn: IGIOPConnection; redo: boolean);
var
  i : integer;
  rec: IIIOPServerInvokeRecord;
  again : Boolean;
begin
  FConnLock.lock();
  try
    conn._AddRef;
    i := FConnections.IndexOf(conn);
    while i >= 0 do
      begin
        FConnections.Delete(i);
        i := FConnections.IndexOf(conn)
      end;
    conn._Release;
  finally
    FConnLock.unlock();
  end; { try/finally }
  conn.terminate();
{$IFDEF USE_IOP_CACHE}
  if FCacheUsed and (FCacheRecord.connection = conn) then
    begin
      FORB.cancel(FCacheRecord.orbid);
      FCacheUsed := false;
    end;
{$ELSE}
  again := true;
  while again do
    begin
      FORBIdsLock.lock();
      try
        again := False;
        for i := 0 to FORBIds.count - 1 do
          begin
            rec := IIIOPServerInvokeRecord(FORBIds.Items(i));
            if rec.active and (rec.connection = conn) then
              begin
                rec.deactivate();
                abort_invoke_orbid(rec);
                again := True;
                break;
              end;
          end;
      finally
        FORBIdsLock.unlock();
      end; { try/finally }
    end;
  deref_connection(conn);
{$ENDIF}
end;

function TIIOPServer.listen: Boolean;
var
  addr : IInetAddress;
begin
  addr := TInetAddress.Create('0.0.0.0',0,STREAM);
  result := listen_addr(addr,nil);
end;

function TIIOPServer.listen_addr(addr, addr2: IAddress): Boolean;
var
  tserv: ITransportServer;
  prof : IIORProfile;
  mc : IMultiComponent;
begin
  result := false;
  tserv := addr.make_transport_server();
  tserv.create_thread();
  if not tserv.bind(addr) then exit;
  tserv.block(FConnMgr.dispatcher().isblocking());
  { TODO -oOVS : dispatcher }
  tserv.aselect(FConnMgr.dispatcher(),self);
  //tserv.aselect(FORB.dispatcher(),self);
  mc := TMultiComponent.Create;
  prof := IIORProfile(tserv.get_address.make_ior_profile('',1,mc));
  FOrb.ior_template.add_profile(prof);
  FTServers.Add(tserv);
  tserv.start();
  result := true;
end;

function TIIOPServer.bind(mid: ORBMessageID; const repoid: string;
  const oid: IObjectTag; const addr: IAddress): Boolean;
begin
  result := false;
end;

function TIIOPServer.invoke(mid: ORBMessageID; const obj: IORBObject;
  const req: IORBRequest; const pr: IPrincipal; response_exp: Boolean): Boolean;
begin
  result := false;
  assert(false,'not implemented');
end;

procedure TIIOPServer.cancel(mid: ORBMessageID);
begin
  assert(false);
end;


//**************************************************************************
// TIIOPServerInvokeRecord
//**************************************************************************
function TIIOPServerInvokeRecord.connection: IGIOPConnection;
begin
  result := FConnection;
end;

procedure TIIOPServerInvokeRecord.clean();
begin
 FConnection := nil;
 FRequest := nil;
 FObject := nil;
 FORBId := nil;
 FREQId := 0;
 FActive := true;
end;

function TIIOPServerInvokeRecord.get_request: IORBRequest;
begin
  result := nil;
  assert(false,'not implemented');
end;

procedure TIIOPServerInvokeRecord.init_invoke(const conn: IGIOPConnection;
  reqid: MessageID; orbid: ORBMessageID; const req: IORBRequest; const obj: IORBObject);
begin
  FConnection:= conn ;
  FORBId := orbid;
  FREQId := reqid;
  FRequest := req;
  FObject:= obj;
  FOrbMsgId := ORB_Instance.get_msgid(orbid);
  FActive := true;
end;

function TIIOPServerInvokeRecord.orbid: ORBMessageID;
begin
  result := FORBId;
end;

function TIIOPServerInvokeRecord.reqid: MessageID;
begin
  result := FReqID;
end;

procedure TIIOPServerInvokeRecord.set_orbid(oid: ORBMessageID);
begin
  //assert(false,'not implemented');
  FORBId := oid;
end;

procedure TIIOPServerInvokeRecord.set_reqid(reqid: MessageID);
begin
  //assert(false,'not implemented');
  FREQId := reqid;
end;

function TIIOPServerInvokeRecord.orbmsgid: MessageID;
begin
  result := FOrbMsgId;
end;

function TIIOPServerInvokeRecord.active: boolean;
begin
  result := FActive;
end;

procedure TIIOPServerInvokeRecord.deactivate;
begin
  FActive := false;
end;

end.
