// **********************************************************************
//
// Copyright (c) 2001 MT Tools.
//
// All Rights Reserved
//
// MT_DORB is based in part on the product DORB,
// written by Shadrin Victor
//
// See Readme.txt for contact information
//
// **********************************************************************
unit ior;

interface

uses
  windows,Types,SysUtils,decoder,classes,address,orbcomp,
  orb_int,orb,code_int,addr_int,iior_int,mcomp_int,iiop_int;

type

  TIORProfile = class;
  TIORProfileDecoder = class;

  {** Implements org.omg.CORBA.Environment }
  TIOR = class(TORBObject,IIOR,IORBObject)
  private
    ObjectID : string;
    FProfiles : IInterfaceList;
  protected
    function decode(const dec: IDecoder): Boolean;
    procedure encode(const enc: IEncoder);
    procedure print(s : TStream);
    procedure clean();
    function get_profile(id : ProfileID): IIORProfile;
    procedure add_profile(const prof : IIORProfile);
    procedure set_objectkey(const key: string; const len: _ulong);
    procedure set_objectid(const oid: string);
    function size: integer;
    function profile(ind: _ulong): IIORProfile;
    function objid: string;
    function stringify: string;
  public
    constructor Create(); overload;
    constructor Create(const tmpl: IIOR); overload;
    constructor Create(const ref: string); overload;
    function from_string(const Value: string): Boolean;
  end;

  TIORProfile = class(TORBObject,IIORProfile)
  private
    FID : ProfileID;
    FAddress : IInetAddress;
    FMultiComponent : IMultiComponent;
  protected
    function address(): IAddress;
    function id(): ProfileID; virtual;
    function objkey(var len: _ulong): string; virtual;
    procedure encode(const enc: IEncoder); virtual; abstract;
    function clone(): IIORProfile; virtual; abstract;
    procedure print(s: TStream); virtual; abstract;
    function objkeylen: _ulong; virtual;
    procedure set_objectkey(const val: string; const len : _ulong); virtual; abstract;
  public
    class procedure register_decoder(tipd : Pointer);
    class procedure unregister_decoder(tipd : Pointer);
    class function decode(dec: IDecoder): IIORProfile;
    class function decode_body(dec: IDecoder;pid: ProfileID;len : _ulong): IIORProfile;
  end;

  TGIOPSimpleProfile = class(TIORProfile,IIORProfile)
  private
    FObjKey : string;
    FOKLen : integer;
  protected
    procedure encode(const enc: IEncoder); override;
    function objkey(var len: _ulong): string; override;
    function clone(): IIORProfile; override;
    procedure print(s: TStream); override;
    procedure set_objectkey(const val: string; const len : _ulong); override;
    function id(): ProfileID; override;
  public
    constructor Create(const objkey: string = '';len: _ulong = 0);
    constructor Create_prof(const prof: IIORProfile);
  end;

  TIIOPProfile = class(TIORProfile,IIORProfile,IIOPProfile)
  private
    FObjKey : string;
    FOKLen : integer;
    FVersion : _ushort;
  protected
    procedure encode(const enc: IEncoder); override;
    function clone(): IIORProfile; override;
    function objkey(var len: _ulong): string; override;
    procedure print(s: TStream); override;
    function comps: IMultiComponent;
    function version: _ushort;
    function objkeylen: _ulong; override;
    procedure set_objectkey(const val: string; const len : _ulong); override;
  public
    constructor Create(const objkey: string;len: _ulong;
        address: IInetAddress;comps: IMultiComponent;
        version : _ushort = $0100; tagid: ProfileID = TAG_INTERNET_IOP);
    constructor Create_prof(const prof: IIOPProfile);
  end;

  TMultiCompProfile  = class(TIORProfile,IIORProfile,IMultiComponentProfile)
  protected
    procedure encode(const enc: IEncoder); override;
    function clone(): IIORProfile; override;
    procedure set_objectkey(const val: string; const len : _ulong); override;
  public
    constructor Create(mc: IMultiComponent; id : ProfileID);
    constructor Create_prof(const prof: IMultiComponentProfile);
    procedure print(s: TStream); override;
    function get_components(): IMultiComponent;
  end;

  TIORProfileDecoder = class(TORBObject,IIORProfileDecoder)
  private
    FID : Cardinal;
  public
    constructor Create(pid : ProfileID);
    destructor Destroy(); override;
    function has_id(id: ProfileID): Boolean;
    function decode(const dec: IDecoder;pid: ProfileID;len : _ulong): IIORProfile; virtual; abstract;
  end;

  TIIOPProfileDecoder = class(TIORProfileDecoder,IIORProfileDecoder)
  public
    function decode(const dec: IDecoder;pid: ProfileID;len : _ulong): IIORProfile; override;
  end;

  TMultiCompProfileDecoder = class(TIORProfileDecoder,IIORProfileDecoder)
  public
    function decode(const dec: IDecoder;pid: ProfileID;len : _ulong): IIORProfile; override;
  end;


implementation

uses
  buffer,pstream,Utils,encoder;

var
  Decoders : TList;
  IIOPProfileDecoder : IIORProfileDecoder;
  MultiCompProfileDecoder : IIORProfileDecoder;

//**********************************************************************
// TIOR
//**********************************************************************
constructor TIOR.Create();
begin
  FProfiles := TInterfaceList.Create();
  ObjectID := '';
end;

constructor TIOR.Create(const ref: string);
begin
  FProfiles := TInterfaceList.Create();
  ObjectID := '';
  from_string(ref);
end;


constructor TIOR.Create(const tmpl: IIOR);
var
  i : integer;
begin
  FProfiles := TInterfaceList.Create();
  for i := 0 to tmpl.size -1 do
    add_profile(tmpl.profile(i).clone);
  ObjectID := tmpl.objid;
end;

function TIOR.objid: string;
begin
  result := ObjectID;
end;

function TIOR.size: integer;
begin
  result := FProfiles.Count
end;

function TIOR.profile(ind: _ulong): IIORProfile;
begin
  result := nil;
  if ind < _ulong(size) then
    result := IIORProfile(FProfiles[ind]);
end;

procedure TIOR.set_objectkey(const key: string; const len: _ulong);
var
  i : integer;
begin
  for i := 0 to Pred(FProfiles.Count) do
    IIORProfile(FProfiles[i]).set_objectkey(key,len);
end;

procedure TIOR.set_objectid(const oid: string);
begin
  ObjectID := oid;
end;

function TIOR.from_string(const Value: string): Boolean;
const
  chars = '0123456789abcdef';
var
  Buffer : IBuffer;
  i,size : integer;
  b_order : TByteOrder;
  dec : IDecoder;
  b,lb,rb : Byte;
  val : string;
begin
  result := false;
  size := Length(Value);
  if (Pos('IOR:',Value) <> 1) or ((size mod 2 and 1) <> 0) then
    exit;
  Buffer := TBuffer.Create(round(size/2));
  val := LowerCase(Value);
  i := 5;
  while i < size do
    begin
      if val[i] = #13 then break;
      if val[i] = #10 then break;
      if val[i] < 'a' then
        lb := Ord(val[i]) - Ord('0')
      else
        lb := Ord(val[i]) - Ord('a') + 10;
      inc(i);
      if val[i] = #13 then break;
      if val[i] = #10 then break;
      if val[i] < 'a' then
        rb := Ord(val[i]) - Ord('0')
      else
        rb := Ord(val[i]) - Ord('a') + 10;
      inc(i);
      Buffer.put(Octet((lb shl 4) or rb))
    end;
  dec := TDecoder.Create(Buffer);
  if dec.get_octet(Octet(b)) then
    begin
      b_order := TByteOrder(b);
      dec.byteorder(b_order);
      if not decode(dec) then
        clean;
    end;
  Result := True;
end;

procedure TIOR.encode(const enc: IEncoder);
var
   i : integer;
   stat : TEncapsState;
begin
  enc.put_string(ObjectId);
  enc.seq_begin(FProfiles.Count);
  for i := 0 to Pred(FProfiles.Count) do
    begin
      enc.put_ulong(IIORProfile(FProfiles[i]).id);
      enc.encaps_begin(stat);
      IIORProfile(FProfiles[i]).encode(enc);
      enc.encaps_end(stat);
    end;
  enc.seq_end;
end;

function TIOR.decode(const dec: IDecoder): Boolean;
var
  len,i : longint;
  prof : IIORProfile;
label
  error;
begin
  result := false;
  clean();
  if not dec.get_string(ObjectID) then goto error;
  //if ObjectID = '' then exit;
  if not dec.get_ulong(_ulong(len)) then goto error;
  for i := 0 to len-1 do
    begin
      prof := TIORProfile.decode(dec);
      if prof = nil then goto error;
      add_profile(prof);
    end;
  result := true;
  exit;
error:
  clean();
end;

procedure TIOR.clean();
begin
  FProfiles.Clear;
end;

procedure TIOR.add_profile(const prof: IIORProfile);
begin
  FProfiles.Add(prof);
end;

procedure TIOR.print(s : TStream);
var
  ps : TPrintStream;
  i : integer;
begin
  ps := TPrintStream(s);
  ps.writestr('Repo ID: ');
  ps.writeln(ObjectID);
  ps.writecrlf;
  for i := 0 to Pred(FProfiles.Count) do
    IIORProfile(FProfiles[i]).print(s);
  ps.writeeol;
end;

function TIOR.stringify: string;
var
  enc: IEncoder;
  buf : IBuffer;
  i : integer;
  o : Octet;
begin
  result := 'IOR:';
  enc := TEncoder.Create();
  enc.put_octet(Octet(enc.byteorder = LittleEndian));
  encode(enc);
  buf := enc.buffer;
  for i := 0 to Pred(buf.length) do
    begin
      buf.get(o);
      result := result + octet_to_xdigit(Octet(Byte(o) shr 4));
      result := result + octet_to_xdigit(Octet(Byte(o) and $F));
    end;
end;

function TIOR.get_profile(id : ProfileID): IIORProfile;
var
  i : integer;
begin
  result := nil;
  for i := 0 to Pred(FProfiles.Count) do
    if (id = TAG_ANY) or (IIORProfile(FProfiles[i]).id() = id) then
      begin
        result := IIORProfile(FProfiles[i]);
        exit;
      end;
end;

//**********************************************************************
// TIORProfile
//**********************************************************************
class function TIORProfile.decode(dec: IDecoder): IIORProfile;
var
  state : TEncapsState;
  pid,len,nextpos : _ulong;
begin
  result := nil;
  if not dec.get_ulong(pid) then exit;
  if not dec.encaps_begin(state,len) then exit;
  nextpos := dec.Buffer.RPosition+len;
  result := decode_body(dec,pid,len);
  dec.Buffer.RseekFromBegin(nextpos);
  if not dec.encaps_end(state) then
    result := nil;
end;

function TIORProfile.id(): ProfileID;
begin
  result := FID;
end;

function TIORProfile.objkeylen: _ulong;
begin
  result := 0
end;

function TIORProfile.objkey(var len: _ulong): string;
begin
  result := '';
  len := 0;
end;

class function TIORProfile.decode_body(dec: IDecoder;pid: ProfileID;len : _ulong): IIORProfile;
var
  i : integer;
begin
  result := nil;
  for i := 0 to Pred(Decoders.Count) do
    if IIORProfileDecoder(Decoders[i]).has_id(pid) then
      break;
  if i <= Pred(Decoders.Count) then
    result := IIORProfileDecoder(Decoders[i]).decode(dec,pid,len);
end;

class procedure TIORProfile.register_decoder(tipd : Pointer);
begin
  Decoders.Add(tipd);
end;

class procedure TIORProfile.unregister_decoder(tipd : Pointer);
begin
  Decoders.remove(tipd);
end;

function TIORProfile.address(): IAddress;
begin
  result := FAddress;
end;


//**********************************************************************
// TIORProfileDecoder
//**********************************************************************
constructor TIORProfileDecoder.Create(pid : ProfileID);
begin
  FID := pid;
  TIORProfile.register_decoder(Pointer(IIORProfileDecoder(self)));
end;

destructor TIORProfileDecoder.Destroy();
begin
  TIORProfile.unregister_decoder(Pointer(IIORProfileDecoder(self)));
end;

function TIORProfileDecoder.has_id(id: ProfileID): Boolean;
begin
  result := FID = id;
end;

//**********************************************************************
// TIIOProfileDecoder
//**********************************************************************
function TIIOPProfileDecoder.decode(const dec: IDecoder;pid: ProfileID;len : _ulong): IIORProfile;
var
  minor,major : byte;
  version : _ushort;
  port : _ushort;
  host,okey : string;
  objkey : PByteArray;
  l : _ulong;
  components : IMultiComponent;
begin
  result := nil;
  if not dec.get_octet(Octet(major)) then exit;
  if not dec.get_octet(Octet(minor)) then exit;
  version := ((major shl 8) or minor);
  if version > $0102 then exit;
  if not dec.get_string(host) then exit;
  if not dec.get_ushort(port) then exit;
  if not dec.seq_begin(l) then exit;
  if not (dec.Buffer.Length >= l) then exit;
  objkey := dec.Buffer.Data();
  if not dec.Buffer.RseekFromCurrent(l) then exit;
  if not dec.seq_end() then exit;
  components := nil;
  if (major > 1) or (minor > 0) then
    begin
      components := TMultiComponent.Create;
      if not components.decode(dec) then
        begin
          components := nil;
          exit;
        end;
    end;
  setlength(okey,l);
  move(objkey^,PChar(okey)^,l);
  result := TIIOPProfile.Create(okey, l,
      TInetAddress.Create(host,port,STREAM),
      components,
      version,
      FID);
end;

//**********************************************************************
// TMultiCompProfileDecoder
//**********************************************************************
function TMultiCompProfileDecoder.decode(const dec: IDecoder;pid: ProfileID;len : _ulong): IIORProfile;
var
  components : IMultiComponent;
begin
  result := nil;
  components := TMultiComponent.Create;
  if not components.decode(dec) then
    begin
      components := nil;
      exit;
    end;
  result := TMultiCompProfile.Create(components,pid);
end;

//**********************************************************************
// TIIOPProfile
//**********************************************************************
constructor TIIOPProfile.Create(const objkey: string;len: _ulong;
   address: IInetAddress;comps: IMultiComponent;
   version : _ushort; tagid: ProfileID);
begin
  inherited Create();
  FOKLen := len;
  FObjKey := objkey;
  FAddress := address;
  self.FMultiComponent := comps;
  FVersion := version;
  if (FVersion < $0101) and (comps <> nil) and (comps.size > 0)  then
    FVersion := $0101;
  FID := tagid;
end;

constructor TIIOPProfile.Create_prof(const prof: IIOPProfile);
var
  objkey: string;
  len: _ulong;
begin
  inherited Create();
  objkey := prof.objkey(len);
  FOKLen := len;
  FObjKey := objkey;
  FAddress := IInetAddress(prof.address);
  self.FMultiComponent := IMultiComponent(prof.comps);
  FVersion := prof.version;
  if (FVersion < $0101) and (comps <> nil) and (comps.size > 0)  then
    FVersion := $0101;
  FID := prof.id;
end;

function TIIOPProfile.version: _ushort;
begin
  result := FVersion;
end;

function TIIOPProfile.objkeylen: _ulong;
begin
  result := FOKLen;
end;


function TIIOPProfile.comps: IMultiComponent;
begin
  result := FMultiComponent;
end;

function TIIOPProfile.clone(): IIORProfile;
begin
  result := TIIOPProfile.Create_prof(self);
end;

procedure TIIOPProfile.set_objectkey(const val: string; const len : _ulong);
begin
  FOKLen := len;
  FObjKey := val;
end;

procedure TIIOPProfile.encode(const enc: IEncoder);
begin
  enc.put_octet(Octet(FVersion shr 8));
  enc.put_octet(Octet(FVersion));
  enc.put_string(FAddress.host);
  enc.put_short(FAddress.Port);
  enc.seq_begin(FOKLen);
  enc.put_octets(PChar(FObjKey)^,FOKLen);
  enc.seq_end;
  if FVersion > $0101 then
    FMultiComponent.encode(enc);
end;

function TIIOPProfile.objkey(var len: _ulong): string;
begin
  result := FObjKey;
  len := FOKLen;
end;

procedure TIIOPProfile.print(s: TStream);
var
  ps : TPrintStream;
begin
  ps := TPrintStream(s);
  ps.writeln('IIOP Profile');
  ps.writeln ('   Version: '+ IntToStr(FVersion shr 8)+'.'+ IntToStr(Fversion and $FF));
  ps.writeln ('   Address: '+ FAddress.toString);
  ps.writestr('  Location: iooploc://');
  if FVersion <> $0100 then
      ps.writeln(IntToStr(FVersion shr 8)+'.'+ IntToStr(Fversion and $FF)+'@');
  ps.writestr (FAddress.Host+':'+IntToStr(FAddress.Port));
  if FOKLen > 0 then
    ps.writestr(url_encode(PChar(FObjKey)));
  ps.writecrlf;
  ps.writecrlf;
end;


//**********************************************************************
// TMultiCompProfile
//**********************************************************************
constructor TMultiCompProfile.Create(mc: IMultiComponent; id : ProfileID);
begin
  FID := id;
  FMultiComponent := mc;
end;

constructor TMultiCompProfile.Create_prof(const prof: IMultiComponentProfile);
begin
  FMultiComponent := prof.get_components;
  FID := prof.id;
end;

procedure TMultiCompProfile.set_objectkey(const val: string; const len : _ulong);
begin

end;

function TMultiCompProfile.clone: IIORProfile;
begin
  result := TMultiCompProfile.Create_prof(self);
end;

procedure TMultiCompProfile.encode(const enc: IEncoder);
begin
  FMultiComponent.encode(enc);
end;

procedure TMultiCompProfile.print(s: TStream);
begin
  TPrintStream(s).writeln('Multiple Components Profile');
  FMultiComponent.print(s);
end;

function TMultiCompProfile.get_components(): IMultiComponent;
begin
  result := FMultiComponent;
end;

//********************************************************************
// TGIOPSimpleProfile
//********************************************************************
constructor TGIOPSimpleProfile.Create(const objkey: string;len: _ulong);
begin
  FOKLen := len;
  FObjKey := objkey;
end;

constructor TGIOPSimpleProfile.Create_prof(const prof: IIORProfile);
begin
  FObjKey := prof.objkey(_ulong(FOKLen));
end;

procedure TGIOPSimpleProfile.encode(const enc: IEncoder);
begin
  assert(false,'not implemented');
end;

function TGIOPSimpleProfile.id(): ProfileID;
begin
  result := 10000;
end;

function TGIOPSimpleProfile.objkey(var len: _ulong): string;
begin
  result := FObjKey;
  len := FOKLen;
end;

function TGIOPSimpleProfile.clone(): IIORProfile;
begin
  result := TGIOPSimpleProfile.Create_prof(self);
end;

procedure TGIOPSimpleProfile.print(s: TStream);
begin
  s.Write('GIOPSimple Profile',20);
end;

procedure TGIOPSimpleProfile.set_objectkey(const val: string; const len : _ulong);
begin
  FOKLen := len;
  FObjKey := val;
end;

initialization
  Decoders := TList.Create;
  IIOPProfileDecoder := TIIOPProfileDecoder.Create(TAG_INTERNET_IOP);
  MultiCompProfileDecoder := TMultiCompProfileDecoder.Create(TAG_MULTIPLE_COMPONENTS);
finalization
  IIOPProfileDecoder := nil;
  MultiCompProfileDecoder := nil;
  Decoders.Free;
end.
