// **********************************************************************
//
// Copyright (c) 2001 MT Tools.
//
// All Rights Reserved
//
// MT_DORB is based in part on the product DORB,
// written by Shadrin Victor
//
// See Readme.txt for contact information
//
// **********************************************************************
unit mtmanager;

interface

uses
  windows, sysutils, classes, contnrs, mtmanager_int, OSThread, types;

type
  TOP_id_type_decl = (odAccept, odReader, odWriter, odDeCode, odORB);
  TMsgChannelTypes = (mcActive, mcPassive, mcDirect);

  // Forward declarations
  TThreadPool = class;
  TThreadPoolManager = class;
  TWorkerThread = class;

  TMsgType = class(TInterfacedObject, IMsgType)
  protected
    FType: TMTMesType;
    FPtr: pointer;
    FValue: integer;
    function get_type: TMTMesType;
    procedure set_type(msgType: TMTMesType);
    function data: pointer;
  public
    constructor Create(ptr: pointer; AType: TMTMesType = mtProcess);
    destructor Destroy; override;
    property Value: integer read FValue write FValue;
  end;

  TMsgChannel = class(TInterfacedObject, IMsgChannel)
  protected
    FThreadPool: IThreadPool;
    procedure setThreadPool(AThreadPool: IThreadPool);
    function getThreadPool: IThreadPool;
    procedure check_msg(); virtual;
    function get_msg(idType: TOP_id_type): IMsgType; virtual;
    procedure put_msg(nextOpId: TOP_id_type; msg: IMsgType); virtual; abstract;
  public
  end;

  TActiveMsgQueue = class(TMsgChannel, IMsgChannel)
  private
    FQueue: TQueue;
    FAccess: TMutex;
  protected
    procedure check_msg(); override;
    procedure put_msg(nextOpId: TOP_id_type; msg: IMsgType); override;
  public
    constructor Create();
    destructor Destroy; override;
  end;

  TPassiveMsgQueue = class(TMsgChannel, IPassiveMsgQueue)
  private
    Fkt: IWorkerThread;
    FQueue: TQueue;
    FAccess: TMutex;
    FMsgCnt: TSemaphore;
  protected
    procedure setWorkerThread(kt: IWorkerThread);
    function get_msg(idType: TOP_id_type): IMsgType; override;
    procedure put_msg(idType: TOP_id_type; msgType: IMsgType); override;
  public
    constructor Create();
    destructor Destroy; override;
  end;

  TDirectMsgConnector = class(TMsgChannel, IDirectMsgConnector)
  private
    FOp: IOperation;
  protected
    procedure setOperation(op: IOperation);
    procedure put_msg(idType: TOP_id_type; msgType: IMsgType); override;
  public
  end;

  TOperation = class(TMsgChannel, IOperation)
  private
  protected
    FInfo: IOperationInfoNode;
    FWThread: IWorkerThread;
    function GetInfo(): IOperationInfoNode;
    procedure SetInfo(value: IOperationInfoNode);
    function GetThread(): IWorkerThread;
    procedure SetThread(value: IWorkerThread);
    procedure start();
    procedure initShutdown(); virtual;
    procedure finalizeShutdown(); virtual;
    procedure send_msg(nextOpId: TOP_id_type; msg: IMsgType);
    procedure process(msg: IMsgType); virtual;
    procedure _run(); virtual; abstract;
    function copy(): IOperation; virtual; abstract;
    procedure put_msg(nextOpId: TOP_id_type; msg: IMsgType); override;
  public
    constructor Create(); overload;
    constructor Create(AThreadPool: IThreadPool); overload;
    constructor Create(op: IOperation); overload;
    destructor Destroy; override;
  end;

  TOperationInfoNode = class(TInterfacedObject, IOperationInfoNode)
  private
    FNextOP: IMsgChannel;
    FTp: IThreadPool;
    FOpID: TOP_id_type;
  protected
    function getThreadPool(): IThreadPool;
    procedure setThreadPool(tp: IThreadPool);
    function get_OP_id(): TOP_id_type;
    procedure set_OP_id(id: TOP_id_type);
    function get_nextOP(): IMsgChannel;
    procedure set_nextOP(nextOP: IMsgChannel);
  end;

  TActiveOperation = class(TOperation, IActiveOperation)
  private
  protected
    FInputMC: IPassiveMsgQueue;
    procedure registerInputMC(mc: IPassiveMsgQueue);
    function getInputMC(): IPassiveMsgQueue;
    procedure initShutdown(); override;
    procedure _run(); override;
  public
    constructor Create(AThreadPool: IThreadPool); overload;
    constructor Create(op: IActiveOperation); overload;
  end;

  TPassiveOperation = class(TOperation, IPassiveOperation)
  private
    FMsgType: IMsgType;
  protected
    procedure put_msg(idType: TOP_id_type; msgType: IMsgType); override;
    procedure _run(); override;
  end;

  TWorkerThread = class(TORBThread, IWorkerThread)
  private
    FRefCount: Integer;
    FTP: IThreadPool;
    FOp: IOperation;
    FState: TWorkerThreadState;
    FOpCritical: TMutex;
    FStateSem: TSemaphore;
    //FMsg: IMsgType;
    {procedure _run_single_queue_input();
    procedure _run_multi_queue_input();
    procedure _run_process();}
    procedure wait_for_state_change();
    function GetState: TWorkerThreadState;
    procedure SetState(const Value: TWorkerThreadState);
  protected
    function QueryInterface(const IID: TGUID; out Obj): HResult; stdcall;
    function _AddRef: Integer; stdcall;
    function _Release: Integer; stdcall;
    procedure setThreadPool(AThreadPool: IThreadPool);
    function getThreadPool: IThreadPool;
    procedure registerOperation(op: IOperation);
    procedure deregisterOperation(op: IOperation);
    procedure post_state_change();
    procedure put_msg(msgType: IMsgType);
    procedure mark_idle();
    procedure mark_busy();
  public
    constructor Create(); overload;
    constructor Create(AThreadPool: IThreadPool); overload;
    destructor Destroy; override;

    function getOperation(): IOperation;
    procedure _run(arg: pointer); override;
  end;

  TWorkerThreadList = class(TList);

  TThreadPool = class(TMsgChannel, IThreadPool)
  private
    FTPLock: TMutex;
    FTpm: IThreadPoolManager;
    FBusyThreads,
    FIdleThreads: TList;
    FOp: IOperation;
    FInputMC: TMsgChannel;
    FMax,
    FMaxIdle,
    FMinIdle,
    FCountAll: cardinal;
  protected
    function get_idle_thread(): IWorkerThread;
    procedure mark_busy(wt: IWorkerThread);
    procedure mark_idle(wt: IWorkerThread);
    function getOperation(): IOperation;
    procedure setThreadPoolManager(tpm: IThreadPoolManager);
  public
    constructor Create(AMax, AMaxIdle, AMinIdle: cardinal);
    destructor Destroy; override;

    function new_idle_thread(): IWorkerThread;
    procedure remove_thread(kt: IWorkerThread);
    // fire up all threads
    procedure start_threads(arg: pointer = nil); virtual;

    function getThreadPoolManager(): IThreadPoolManager;

    procedure setOperation(op: IOperation);

    procedure registerInputMC(mc: TMsgChannel);
    function getInputMC(): TMsgChannel;

    procedure put_msg(nextOpId: TOP_id_type; msgType: IMsgType); override;
  end;

  TThreadPoolManager = class(TMsgChannel, IThreadPoolManager)
  private
    FTp: TStringList;
    FUserCnt: _ulong;
  public
    constructor Create();
    destructor Destroy; override;

    function register_new_tp(tp: IThreadPool): _ulong;
    procedure register_tp(const opId: TOP_id_type; tp: IThreadPool);

    function _getThreadPool(const opId: TOP_id_type): IThreadPool;

    function get_idle_thread(const opId: TOP_id_type): IWorkerThread;

    function getOperation(opId: TOP_id_type): IOperation;

    procedure put_msg(nextOpId: TOP_id_type; msgType: IMsgType); override;
  end;

  TMTDispatcher = class(TPassiveOperation, IPassiveOperation)
  protected
    procedure process(msg: IMsgType); override;
    function copy(): IOperation; override;
  public
    constructor Create(); overload;
    constructor Create(op: IOperation); overload;
  end;

function TPM(): IThreadPoolManager;
procedure thread_setup();

implementation

uses
  iop,giop_int;

type
  TTm_Init_T = record
    OP_type: TOP_id_type_decl;
    max: cardinal;
    max_idle: cardinal;
    min_idle: cardinal;
    mct: TMsgChannelTypes;
    mc_size: cardinal;
    op: IOperation;
    start: boolean;
  end;

const
  tmInitSize = 5;
  tmInit: array [0..tmInitSize-1] of TTm_Init_T =
   ((OP_type: odAccept; max: 20; max_idle: 0; min_idle: 0; mct: mcPassive; mc_size: 0; op: nil; start: FALSE),
    (OP_type: odWriter; max: 20; max_idle: 0; min_idle: 0; mct: mcPassive; mc_size: 0; op: nil; start: FALSE),
    (OP_type: odReader; max: 20; max_idle: 0; min_idle: 0; mct: mcPassive; mc_size: 0; op: nil; start: FALSE),
    (OP_type: odDeCode; max: 20; max_idle: 0; min_idle: 0; mct: mcActive;  mc_size: 1; op: nil; start: TRUE),
    (OP_type: odORB;    max: 10; max_idle: 0; min_idle: 0; mct: mcActive;  mc_size: 1; op: nil; start: TRUE));

var
  tpmVar: IThreadPoolManager;

function TPM(): IThreadPoolManager;
begin
  result := tpmVar;
end;

procedure thread_setup();
var
  i: integer;
  tpmObj: TThreadPoolManager;
  tpObj: TThreadPool;
  tp: IThreadPool;
  mq: TDirectMsgConnector;
begin
  tpmObj := TThreadPoolManager.Create;
  for i := 0 to tmInitSize-1 do begin
    tpObj := TThreadPool.Create(tmInit[i].max, tmInit[i].max_idle, tmInit[i].min_idle);
    tp := tpObj;
    tpmObj.register_tp(ord(tmInit[i].OP_type), tp);
    case tmInit[i].OP_type of
      odDeCode: tpObj.setOperation(TInputHandler.Create() as IPassiveOperation);
      odORB: tpObj.setOperation(TMTDispatcher.Create() as IPassiveOperation);
    end;
    case tmInit[i].mct of
      mcActive:
        if tmInit[i].mc_size <> 0 then tpObj.registerInputMC(TActiveMsgQueue.Create());
      mcDirect:
        begin
          mq := TDirectMsgConnector.Create;
          mq.setOperation(tpObj.getOperation());
          tpObj.registerInputMC(mq);
        end;
    end;
    if tmInit[i].start then tpObj.start_threads();
  end;
  tpmVar := tpmObj;
end;

{ TOperationInfoNode }

function TOperationInfoNode.get_nextOP: IMsgChannel;
begin
  result := FNextOP;
end;

function TOperationInfoNode.get_OP_id: TOP_id_type;
begin
  result := FOpID;
end;

function TOperationInfoNode.getThreadPool: IThreadPool;
begin
  result := FTp;
end;

procedure TOperationInfoNode.set_nextOP(nextOP: IMsgChannel);
begin
  FNextOP := nextOP;
end;

procedure TOperationInfoNode.set_OP_id(id: TOP_id_type);
begin
  FOpID := id;
end;

procedure TOperationInfoNode.setThreadPool(tp: IThreadPool);
begin
  FTp := tp;
end;

{ TMsgChannel }

procedure TMsgChannel.check_msg;
begin

end;

function TMsgChannel.get_msg(idType: TOP_id_type): IMsgType;
begin
  Assert(false, 'Not implemented');
  result := nil;
end;

function TMsgChannel.getThreadPool: IThreadPool;
begin
  result := FThreadPool;
end;

procedure TMsgChannel.setThreadPool(AThreadPool: IThreadPool);
begin
  FThreadPool := AThreadPool;
end;

{ TOperation }

constructor TOperation.Create(AThreadPool: IThreadPool);
begin
  FInfo := TOperationInfoNode.Create();
  FInfo.setThreadPool(AThreadPool);
end;

constructor TOperation.Create;
begin
  FInfo := TOperationInfoNode.Create();
end;

constructor TOperation.Create(op: IOperation);
begin
  FInfo := op.info;
end;

destructor TOperation.Destroy;
begin
  FInfo := nil;
  inherited;
end;

procedure TOperation.finalizeShutdown;
begin
  // make sure the thread has left the context of
  // this OP, deregisterOP will take care of that !
  if FWThread <> nil then FWThread.deregisterOperation(Self);
end;

function TOperation.GetInfo: IOperationInfoNode;
begin
  result := FInfo;
end;

function TOperation.GetThread: IWorkerThread;
begin
  result := FWThread;
end;

procedure TOperation.initShutdown;
begin

end;

// the assert's below mark methods that have to be overloaded
// in derived classes
// the derived classes have to overload only one of those !!!
procedure TOperation.process(msg: IMsgType);
begin
  Assert(false, 'Implementation error - this should never happen');
end;

procedure TOperation.put_msg(nextOpId: TOP_id_type; msg: IMsgType);
begin
  Assert(false, 'not implemented');
end;

procedure TOperation.send_msg(nextOpId: TOP_id_type; msg: IMsgType);
begin
  FInfo.get_nextOP().put_msg(nextOpId, msg);
end;

procedure TOperation.SetInfo(value: IOperationInfoNode);
begin
  FInfo := value;
end;

procedure TOperation.SetThread(value: IWorkerThread);
begin
  FWThread := value;
end;

procedure TOperation.start;
begin
  FWThread.post_state_change();
end;

{ TActiveOperation }

procedure TActiveOperation._run;
var
  msg: IMsgType;
begin
  while true do begin
    msg := getInputMC().get_msg(FInfo.get_OP_id());
    // We are guaranteed that msg is not NULL because get_msg() asserts
    // this, so don't bother checking it again after the call to get_msg()
    // above!!!
    if msg.get_type() = mtTerminate then begin
      msg := nil;
      Exit;
    end;
    process(msg);
  end; { while }
end;

constructor TActiveOperation.Create(AThreadPool: IThreadPool);
begin
  inherited Create(AThreadPool);
end;

constructor TActiveOperation.Create(op: IActiveOperation);
begin
  inherited Create(op);
  FInputMC := op.getInputMC();
end;

function TActiveOperation.getInputMC: IPassiveMsgQueue;
begin
  result := FInputMC;
end;

procedure TActiveOperation.initShutdown;
var
  msg: IMsgType;
begin
  if getInputMC() <> nil then begin
    msg := TMsgType.Create(nil, mtTerminate);
    getInputMC().put_msg(FInfo.get_OP_id(), msg)
  end;
end;

procedure TActiveOperation.registerInputMC(mc: IPassiveMsgQueue);
begin
  FInputMC := mc;
  FInputMC.setWorkerThread(FWThread);
end;

  { TPassiveOperation }

procedure TPassiveOperation._run;
begin
  if (FMsgType = nil) or (FMsgType.get_type() = mtTerminate) then begin
    FMsgType := nil;
    Exit;
  end;
  process(FMsgType);
  FMsgType := nil;
end;

procedure TPassiveOperation.put_msg(idType: TOP_id_type;
  msgType: IMsgType);
begin
  Assert(Finfo.get_OP_id = idType);
  FMsgType := msgType;
end;

  { TMsgType }

constructor TMsgType.Create(ptr: pointer; AType: TMTMesType);
begin
  FPtr := ptr;
  FType := AType;
end;

function TMsgType.data: pointer;
begin
  result := FPtr;
end;

destructor TMsgType.Destroy;
begin
  inherited;
end;

function TMsgType.get_type: TMTMesType;
begin
  result := FType;
end;

procedure TMsgType.set_type(msgType: TMTMesType);
begin
  FType := msgType;
end;

  { TActiveMsgQueue }

procedure TActiveMsgQueue.check_msg;
var
  kt: IWorkerThread;
  msg: IMsgType;
begin
  FAccess.lock();
  try
    if FQueue.Count <> 0 then begin
      kt := FThreadPool.get_idle_thread();
      if kt = nil then Exit;
      msg := IMsgType(FQueue.Peek);
      Assert(msg <> nil);
      FQueue.Pop;
      kt.put_msg(msg);
      kt.mark_busy();
      kt.post_state_change();
    end;
  finally
    FAccess.unlock();
  end; { try/finally }
end;

constructor TActiveMsgQueue.Create;
begin
  FQueue := TQueue.Create;
  FAccess := TMutex.Create;
end;

destructor TActiveMsgQueue.Destroy;
begin
  FAccess.Free;
  FQueue.Free;
  inherited;
end;

procedure TActiveMsgQueue.put_msg(nextOpId: TOP_id_type; msg: IMsgType);
var
  kt: IWorkerThread;
begin
  Assert(msg <> nil);
  kt := FThreadPool.get_idle_thread();
  if kt <> nil then begin
    kt.put_msg(msg);
    kt.mark_busy;
    kt.post_state_change;
  end
  else begin
    FAccess.lock();
    try
      msg._AddRef;
      FQueue.Push(Pointer(msg));
    finally
      FAccess.unlock();
    end; { try/finally }
  end;
end;

  { TPassiveMsgQueue }

constructor TPassiveMsgQueue.Create;
begin
  FQueue := TQueue.Create;
  FAccess := TMutex.Create;
  FMsgCnt := TSemaphore.Create;
end;

destructor TPassiveMsgQueue.Destroy;
begin
  FMsgCnt.Free;
  FAccess.Free;
  FQueue.Free;
  inherited;
end;

function TPassiveMsgQueue.get_msg(idType: TOP_id_type): IMsgType;
begin
  FMsgCnt.wait(); // wait for a message (block)
  FAccess.lock();
  try
    result := IMsgType(FQueue.Peek);
    Assert(result <> nil);
    result._Release;
    FQueue.Pop;
  finally
    FAccess.unlock();
  end; { try/finally }
end;

procedure TPassiveMsgQueue.put_msg(idType: TOP_id_type; msgType: IMsgType);
begin
  Assert(msgType <> nil);
  FAccess.lock();
  try
    msgType._AddRef;
    FQueue.Push(Pointer(msgType));
  finally
    FAccess.unlock();
  end; { try/finally }
  FMsgCnt.post();
end;

procedure TPassiveMsgQueue.setWorkerThread(kt: IWorkerThread);
begin
  FKt := kt;
end;

{ TDirectMsgConnector }

procedure TDirectMsgConnector.put_msg(idType: TOP_id_type;
  msgType: IMsgType);
var
  oper: IOperation;
begin
  Assert(msgType <> nil);
  oper := FOp.copy();
  oper.put_msg(idType, msgType);
  oper._run();
  oper := nil;
end;

procedure TDirectMsgConnector.setOperation(op: IOperation);
begin
  FOp := op;
end;

{ TWorkerThread }

procedure TWorkerThread._run(arg: pointer);
var
  op: IOperation;
begin
  repeat
    wait_for_state_change();
    if FState = wtsBusy then begin
      FOpCritical.lock();
      //try
        try
          op := getOperation();
          if op <> nil then op._run();
          op := nil;
        finally
          FOpCritical.unlock();
        end; { try/finally }
      {except
        on E: Exception do
          WriteLn(E.Message);
      end;}
      mark_idle();
      post_state_change();
    end;
  until FState = wtsStop;
end;

{procedure TWorkerThread._run_multi_queue_input;
begin

end;

procedure TWorkerThread._run_process;
begin

end;

procedure TWorkerThread._run_single_queue_input;
begin

end;}

constructor TWorkerThread.Create;
begin
  inherited Create();
  FOpCritical := TMutex.Create(false, maRecursive);
  FStateSem := TSemaphore.Create();
end;

constructor TWorkerThread.Create(AThreadPool: IThreadPool);
begin
  Create();
  FTP := AThreadPool;
end;

procedure TWorkerThread.deregisterOperation(op: IOperation);
begin
  FOpCritical.lock();
  try
    if op = FOp then begin
      FOp.Thread := nil;
      FOp := nil;
    end;
  finally
    FOpCritical.unlock();
  end; { try/finally }
end;

destructor TWorkerThread.Destroy;
begin
  inherited;

end;

function TWorkerThread.getOperation: IOperation;
begin
  result := FOp;
end;

function TWorkerThread.GetState: TWorkerThreadState;
begin
  result := FState;
end;

function TWorkerThread.getThreadPool: IThreadPool;
begin
  result := FTP;
end;

procedure TWorkerThread.mark_busy;
begin
  if getThreadPool() <> nil then getThreadPool().mark_busy(Self);
end;

procedure TWorkerThread.mark_idle;
begin
  if getThreadPool() <> nil then getThreadPool().mark_idle(Self);
end;

procedure TWorkerThread.post_state_change;
begin
  FStateSem.post();
end;

procedure TWorkerThread.put_msg(msgType: IMsgType);
begin
  FOp.put_msg(FOp.Info.get_OP_id(), msgType)
end;

procedure TWorkerThread.registerOperation(op: IOperation);
begin
  FOpCritical.lock();
  try
    FOp := op;
    FOp.Thread := Self;
  finally
    FOpCritical.unlock();
  end; { try/finally }
end;

procedure TWorkerThread.SetState(const Value: TWorkerThreadState);
begin
  FState := Value;
end;

procedure TWorkerThread.setThreadPool(AThreadPool: IThreadPool);
begin
  FTP := AThreadPool;
end;

procedure TWorkerThread.wait_for_state_change;
begin
  FStateSem.wait();
end;

function TWorkerThread._AddRef: Integer;
begin
  Result := InterlockedIncrement(FRefCount);
end;

function TWorkerThread._Release: Integer;
begin
  Result := InterlockedDecrement(FRefCount);
  if Result = 0 then
    Destroy;
end;

function TWorkerThread.QueryInterface(const IID: TGUID; out Obj): HResult;
const
  E_NOINTERFACE = HResult($80004002);
begin
  if GetInterface(IID, Obj) then Result := 0 else Result := E_NOINTERFACE;
end;

{ TThreadPool }

constructor TThreadPool.Create(AMax, AMaxIdle, AMinIdle: cardinal);
begin
  FTPLock := TMutex.Create;
  FBusyThreads := TList.Create;
  FIdleThreads := TList.Create;
  FMax := AMax;
  FMaxIdle := AMaxIdle;
  FMinIdle := AMinIdle;
end;

destructor TThreadPool.Destroy;
var
  i: integer;
begin
  //terminate all threads
  for i := 0 to FIdleThreads.Count - 1 do
    IWorkerThread(FIdleThreads.Items[i]).Terminate(nil);
  for i := 0 to FBusyThreads.Count - 1 do
    IWorkerThread(FBusyThreads.Items[i]).Terminate(nil);

  FIdleThreads.Free;
  FBusyThreads.Free;
  FTPLock.Free;
  inherited;
end;

// returns - a idle thread or creates a new thread if needed
// - or NIL if limits are exhausted
// removes threads from the idle list
function TThreadPool.get_idle_thread: IWorkerThread;
begin
  FTPLock.lock();
  try
    // no idle threads
    if FIdleThreads.Count = 0 then begin
      // limit exhausted
      if FCountAll >= FMax then Exit;
      result := TWorkerThread.Create(Self);
      if FOp <> nil then result.registerOperation(FOp.copy());
      FCountAll := Succ(FCountAll);
      result.Start();
    end
    else begin
      result := IWorkerThread(FIdleThreads.Items[0]);
      FIdleThreads.Delete(0);
    end;
    result.State := wtsTransition;
  finally
    FTPLock.unlock();
  end; { try/finally }
end;

function TThreadPool.getInputMC: TMsgChannel;
begin
  result := FInputMC;
end;

function TThreadPool.getOperation: IOperation;
begin
  result := FOp;
end;

function TThreadPool.getThreadPoolManager: IThreadPoolManager;
begin
  result := FTpm;
end;

function TThreadPool.new_idle_thread: IWorkerThread;
begin
  FTPLock.lock();
  try
    result := TWorkerThread.Create(Self);
    result._AddRef;
    result.no := FIdleThreads.Add(Pointer(result));
    result.State := wtsIdle;
    FCountAll := Succ(FCountAll);
    result.start();
  finally
    FTPLock.unlock();
  end; { try/finally }
end;

procedure TThreadPool.put_msg(nextOpId: TOP_id_type; msgType: IMsgType);
begin
  //FOp.put_msg(nextOpId, msgType);
  if FInputMC <> nil then
    FInputMC.put_msg(nextOpId, msgType);
end;

procedure TThreadPool.registerInputMC(mc: TMsgChannel);
begin
  FInputMC := mc;
  FInputMC.setThreadPool(Self);
end;

procedure TThreadPool.remove_thread(kt: IWorkerThread);
begin
  FTPLock.lock();
  try
    case kt.State of
      wtsIdle: FIdleThreads.Remove(Pointer(kt));//Delete(kt.no);
      wtsBusy: FBusyThreads.Remove(Pointer(kt));//Delete(kt.no)
    end;
    FCountAll := Pred(FCountAll);
  finally
    FTPLock.unlock();
  end; { try/finally }
end;

procedure TThreadPool.setOperation(op: IOperation);
begin
  FOp := op;
end;

procedure TThreadPool.setThreadPoolManager(tpm: IThreadPoolManager);
begin
  FTpm := tpm;
end;

// fire up all threads
procedure TThreadPool.start_threads(arg: pointer);
var
  kt: IWorkerThread;
  i: integer;
begin
  for i := 1 to FMinIdle do begin
    kt := new_idle_thread();
    if getOperation() <> nil then kt.registerOperation(getOperation().copy());
  end;
end;

procedure TThreadPool.mark_busy(wt: IWorkerThread);
begin
  FTPLock.lock();
  try
    if wt.State = wtsIdle then FIdleThreads.Remove(Pointer(wt))// Delete(wt.no);
    else wt._AddRef;
    wt.no := FBusyThreads.Add(Pointer(wt));
    wt.State := wtsBusy;                      
  finally
    FTPLock.unlock();
  end; { try/finally }
end;

procedure TThreadPool.mark_idle(wt: IWorkerThread);
begin
  FTPLock.lock();
  try
    if wt.State = wtsBusy then FBusyThreads.Remove(Pointer(wt))// Delete(wt.no);
    else wt._AddRef;
    wt.no := FIdleThreads.Add(Pointer(wt));
    wt.State := wtsIdle;
    if FInputMC <> nil then FInputMC.check_msg();
  finally
    FTPLock.unlock();
  end; { try/finally }
end;

{ TThreadPoolManager }

constructor TThreadPoolManager.Create();
begin
  FUserCnt := 64;
  FTp := TStringList.Create;
end;

destructor TThreadPoolManager.Destroy;
begin
  FTp.Free;
  inherited;
end;

function TThreadPoolManager.get_idle_thread(
  const opId: TOP_id_type): IWorkerThread;
begin
  result := _getThreadPool(opId).get_idle_thread();
end;

function TThreadPoolManager.getOperation(opId: TOP_id_type): IOperation;
begin
  result := _getThreadPool(opId).getOperation();
end;

function TThreadPoolManager._getThreadPool(
  const opId: TOP_id_type): IThreadPool;
begin
  result := IThreadPool(Pointer(FTp.Objects[FTp.IndexOf(IntToStr(opId))]));
end;

procedure TThreadPoolManager.put_msg(nextOpId: TOP_id_type;
  msgType: IMsgType);
begin
  _getThreadPool(nextOpId).put_msg(nextOpId, msgType);
end;

function TThreadPoolManager.register_new_tp(tp: IThreadPool): _ulong;
begin
  tp._AddRef;
  FTp.AddObject(IntToStr(FUserCnt), Pointer(tp));
  tp.setThreadPoolManager(Self as IThreadPoolManager);
  FUserCnt := Succ(FUserCnt);
  result := FUserCnt;
end;

procedure TThreadPoolManager.register_tp(const opId: TOP_id_type;
  tp: IThreadPool);
var
  idx: integer;
begin
  idx := FTp.IndexOf(IntToStr(opId));
  tp._AddRef;
  if idx <> -1 then begin
    IUnknown(Pointer(FTp.Objects[idx]))._Release;
    FTp.Objects[idx] := Pointer(tp);
  end
  else FTp.AddObject(IntToStr(opId), Pointer(tp));
end;

{ TMTDispatcher }

function TMTDispatcher.copy: IOperation;
begin
  result := TMTDispatcher.Create(Self);
end;

constructor TMTDispatcher.Create;
begin
  inherited;
  FInfo.set_OP_id(ord(odORB));
end;

constructor TMTDispatcher.Create(op: IOperation);
begin
  inherited;
  FInfo.set_OP_id(ord(odORB));
end;

procedure TMTDispatcher.process(msg: IMsgType);
var
  msgRec: IOrbMsg;
begin
  msgRec := IOrbMsg(msg.Data());
  case msgRec.orbEvent() of
    omeKillConn:
      begin
        Assert(msgRec.connection().state = esTerminated);
      end;
    omeCloseConn: begin
      msgRec.callback().giop_callback(msgRec.connection(), Closed);
    end;
  end;
  msgRec._Release();
end;

end.
