// **********************************************************************
//
// Copyright (c) 2001 MT Tools.
//
// All Rights Reserved
//
// MT_DORB is based in part on the product DORB,
// written by Shadrin Victor
//
// See Readme.txt for contact information
//
// **********************************************************************
{**
   Implementation of Transport and TransportServer.
 }
unit transp;

interface

{$I ..\dorb.inc}

uses
  types,orb,classes,addr_int,disp_int,Windows,winsock,code_int,orb_int,
  env_int,mtmanager_int,mtmanager;

type

  TPrincipal = class(TInterfacedObject,IPrincipal)
  private
    FTransport: ITransport;
    FInfo : OctetSeq;
  protected
    procedure encode(const enc: IEncoder);
    function decode(const dec: IDecoder): Boolean;
  public
    constructor Create(transp: ITransport = nil);
    class procedure set_peer_info(info: OctetSeq);
  end;

  TTCPTransport = class(TInterfacedObject,ITransport,IORBDispCallBack)
  protected
    FSocket : TSocket;
    FIsBlocking : Boolean;
    FEOF : Boolean;
    FIsBuffering : Boolean;
    FErrors : TList;
    FReadDispatcher,FWriteDispatcher : IORBDispatcher;
    FReadCallBack,FWriteCallback : ITransportCallBack;
    FPeerAddress : IAddress;
    FState: TTransportState;
  public
    procedure block(val: Boolean);
    function connect(const addr : IAddress): Boolean; stdcall;
    procedure open(fd: Longint); stdcall;
    procedure close(); stdcall;
    function write(const buffer;size: _ulong): integer; stdcall;
    function eof(): Boolean; stdcall;
    function bad(): Boolean; stdcall;
    function read(var buffer;size: _ulong): integer; stdcall;
    procedure callback(const disp: IORBDispatcher;event: TDispEvent);
    procedure rselect(const disp: IORBDispatcher;const cb: ITransportCallback); stdcall;
    procedure wselect(const disp: IORBDispatcher;const cb: ITransportCallback); stdcall;
    function bufread(buf: IBuffer;size: _ulong): integer; stdcall;
    function peer(): IAddress;
    function get_principal(): IPrincipal;
  public
    constructor CreateSock(s : TSocket);
    constructor Create();
    destructor Destroy(); override;
  end;

{$IFDEF HAVE_THREADS}
  TTCPTransportServer = class;

  TTransportServerOperationImpl = class(TOperation, IOperation)
  private
    FCb : ITransportServerCallback;
    FTranServ: ITransportServer;
  protected
    procedure _run(); override;
    function copy(): IOperation; override;
  end;
{$ENDIF}

  TTCPTransportServer = class(TInterfacedObject,ITransportServer,{$IFDEF HAVE_THREADS}IOperation,{$ENDIF}IORBDispCallBack)
  private
    FLocalAddress : IAddress;
    FSocket : TSocket;
    FIsBlocking : Boolean;
    FListening : Boolean;
    FDisp : IORBDispatcher;
    FCb : ITransportServerCallback;
    FOperationObj: TTransportServerOperationImpl;
    FOperation: IOperation;
  protected
    function accept(): ITransport;
    procedure aselect(const disp: IORBDispatcher;const cb: ITransportServerCallback);
    function get_address: IAddress;
    function bind(const addr: IAddress): Boolean;
    procedure callback(const disp: IORBDispatcher;event: TDispEvent);
    procedure listen();
    procedure block(doblock: boolean);
{$IFDEF HAVE_THREADS}
    procedure create_thread();
    procedure start();
    property Operation: IOperation read FOperation implements IOperation;
{$ENDIF}
  public
    constructor Create();
  end;

  procedure sockblock(s: TSocket; val: Boolean);
  procedure sockndelay(s: TSocket; val: Boolean);
  procedure sockclose(s: TSocket);
  function sockwrite(s: TSocket;var Buffer; len: _ulong): integer;
  function sockread(s: TSocket;var Buffer; len: _ulong): integer;
  procedure sockreuse(s: TSocket; val: Boolean);
  procedure seterrno();

function errno(): long;

implementation

uses
  address;

threadvar
  errno_var : long;
  ret: long;
  _peer_info: OctetSeq;

function errno(): long;
begin
  result := errno_var;
end;

procedure sockblock(s: TSocket; val: Boolean);
var
  tmp: integer;
begin
  if val then
    tmp := 0
  else
    tmp := 1;
  Assert(winsock.ioctlsocket(s,FIONBIO,tmp) = 0);
end;

procedure sockreuse(s: TSocket; val: Boolean);
var
  _val : longint;
begin
  if val then
    _val := 1
  else
    _val := 0;
  assert(setsockopt (s, SOL_SOCKET, SO_REUSEADDR, PChar(@_val), sizeof(_val))=0);
end;


procedure sockndelay(s: TSocket; val: Boolean);
{$IFDEF TCP_NODELAY}
var
  tmp: integer;
{$ENDIF}
begin
{$IFDEF TCP_NODELAY}
  tmp := Integer(val);
  ret := winsock.setsockopt(s, IPPROTO_TCP, TCP_NODELAY, PChar(@tmp), sizeof(tmp));
  if ret = 0 then exit;
  seterrno();
{$ENDIF}
end;

procedure sockclose(s: TSocket);
begin
  WSASetLastError(0);
  ret := winsock.closesocket(s);
  if ret = 0 then exit;
  seterrno();
end;

function sockread(s: TSocket;var Buffer; len: _ulong): longint;
begin
  WSASetLastError(0);
  result := winsock.recv(s,buffer,len,0);
  if result <> SOCKET_ERROR then
    exit;
  seterrno();
end;

function sockwrite(s: TSocket;var Buffer; len: _ulong): longint;
begin
  WSASetLastError(0);
  result := winsock.send(s,buffer,len,0);
  if result <> SOCKET_ERROR then
    exit;
  seterrno();
end;

constructor TTCPTransport.Create();
begin
  inherited Create();
  FErrors := TList.Create;
  FPeerAddress := TInetAddress.Create('0.0.0.0',0,STREAM);
  FState := tsClosed;
end;

constructor TTCPTransport.CreateSock(s : TSocket);
begin
  Create();
  FSocket := s;
  if FSocket < 0 then
    FSocket := socket(PF_INET,SOCK_STREAM, 0);
  sockblock(FSocket,true);
  sockndelay(FSocket,true);
  FIsBlocking := True;
  FIsBuffering := False;
  FEof := FALSE;
  //FErrors := TList.Create;
  FState := tsOpen;
end;

destructor TTCPTransport.Destroy();
begin
  if assigned(FReadDispatcher) and Assigned(FReadCallBack) then
    begin
      FReadDispatcher.remove(self,_Read);
      FReadDispatcher := nil;
      FReadCallBack.callback(self,tcRemove);
    end;
  if assigned(FWriteDispatcher) and Assigned(FWriteCallBack) then
    begin
      FWriteDispatcher.remove(self,_Write);
      FWriteDispatcher := nil;
      FWriteCallback.callback(self,tcRemove);
    end;
  sockclose(FSocket);
  //close();
  FErrors.Free;
  inherited Destroy();
end;

procedure TTCPTransport.wselect(const disp: IORBDispatcher;const cb: ITransportCallback); stdcall;
begin
  if (FWriteDispatcher <> nil) and (FWriteCallback <> nil) then
    begin
      FWriteDispatcher.remove(self,Types._Write);
      FWriteDispatcher := nil;
      FWriteCallback := nil;
    end;
  if cb <> nil then
    begin
      disp.socketevent(self,FSocket,Types._Write);
      FWriteDispatcher := disp;
      FWriteCallback := cb;
    end;
end;


procedure TTCPTransport.rselect(const disp: IORBDispatcher;const cb: ITransportCallback);
begin
  if (FReadDispatcher <> nil) and (FReadCallback <> nil) then
    begin
      FReadDispatcher.remove(self,Types._Read);
      FReadDispatcher := nil;
      FReadCallback := nil;
    end;
  if cb <> nil then
    begin
      disp.socketevent(self,FSocket,Types._Read);
      FReadDispatcher := disp;
      FReadCallback := cb;
    end;
end;

function TTCPTransport.peer(): IAddress;
var
  sa : TSockAddrIn;
  r : longint;
begin
  result := nil;
  r := sizeof(sa);
  r := getpeername(FSocket,sa,r);
  if r < 0 then exit;
  FPeerAddress.set_socketaddress(sa);
  result := FPeerAddress;
end;

procedure TTCPTransport.block(val: Boolean);
begin
  if FIsBlocking <> val then
    begin
      FIsBlocking := val;
      sockblock(FSocket,val);
    end;
end;

procedure TTCPTransport.close();
begin
  if FSocket >= 0 then sockclose(FSocket);
  FSocket := socket(PF_INET,SOCK_STREAM, 0);
  assert(FSocket >= 0);
  sockndelay(FSocket,True);
  if assigned(FReadDispatcher) and Assigned(FReadCallBack) then
    FReadDispatcher.remove(self,_Read);
  if assigned(FWriteDispatcher) and Assigned(FWriteCallBack) then
    FWriteDispatcher.remove(self,_Write);
  FIsBuffering := false;
  FIsBlocking := true;
  FReadDispatcher := nil;
  FWriteDispatcher := nil;
  FReadCallBack := nil;
  FWriteCallback := nil;
  {if FSocket >= 0 then
    sockclose(FSocket);}
  FErrors.Clear;
  FEOF := false;
  FState := tsClosed;
end;

procedure TTCPTransport.open(fd: Integer);
begin
  Assert(FState = tsClosed);
  if fd < 0 then
    fd := socket(PF_INET,SOCK_STREAM, 0);
  sockndelay(fd,true);
  FSocket := fd;
  FIsBlocking := True;
  FIsBuffering := False;
  FEof := FALSE;
  FState := tsOpen;
end;

function TTCPTransport.connect(const addr : IAddress): Boolean;
var
  sa : TSockAddrIn;
  r : longint;
begin
  sa := addr.socketaddress();
  r := WinSock.connect(FSocket,sa, sizeof (sa));
  result :=  r >= 0;
end;

function TTCPTransport.write(const buffer;size: _ulong): integer;
var
  written : longint;
  temp : Pointer;
begin
  result := 0;
  temp := @buffer;
  while size > 0 do
    begin
      written := sockwrite(FSocket,temp^,size);
      if written <= 0 then
        break
      else
        begin
          inc(result,written);
          temp := Pointer(Longint(@buffer)+result);
          dec(size,written);
        end;
    end;
end;

function TTCPTransport.eof(): Boolean;
begin
  result := FEOF;
end;

function TTCPTransport.bad(): Boolean;
begin
  result := FErrors.Count > 0;
end;

procedure TTCPTransport.callback(const disp: IORBDispatcher;event: TDispEvent);
begin
  case event of
    types._Read : FReadCallback.callback(self,types.tcRead);
    types._Write : FWriteCallback.callback(self,types.tcWrite);
    types._Remove : begin
        FReadCallback := nil;
        FWriteCallback := nil;
        FReadDispatcher := nil;
        FWriteDispatcher := nil;
      end;
    types._Moved : begin
        FWriteDispatcher := disp;
        FReadDispatcher := disp;
      end;
    else
      assert(false,'not implemented');
  end;
end;

function TTCPTransport.read(var buffer;size: _ulong): integer;
var
  r,todo : longint;
  buf : Pointer;
begin
  todo := size;
  buf := @buffer;
  while todo > 0 do
    begin
      r := sockread(FSocket, buf^, todo);
      if r = SOCKET_ERROR then
        begin
          //seterrno();
          if errno_var = EINTR then
            continue;
          if (errno_var = 0) or
             (errno_var = EWOULDBLOCK) or
             (errno_var = EAGAIN) or
             (todo <> longint(size)) then
            break;
          result := r;
          exit;
        end
      else if r = 0 then
        begin
          Feof := true;
          break;
        end
      else
        begin
          dec(todo,r);
          buf := Pointer(Integer(buf)+ r);
        end;
    end;
  result := longint(size) - todo;
end;

function TTCPTransport.bufread(buf: IBuffer;size: _ulong): integer;
var
  i : integer;
begin
  buf.resize(size);
  i := read(Pointer(_ulong(buf.buffer)+buf.WPosition)^,size);
  if i > 0 then
    buf.WseekFromCurrent(i);
  result := i;
end;

function TTCPTransport.get_principal: IPrincipal;
begin
  result := TPrincipal.Create(self);
end;



//********************************************************************
// TTCPTransportServer
//********************************************************************
constructor TTCPTransportServer.Create();
begin
  FLocalAddress := TInetAddress.Create('0.0.0.0',0,STREAM);
  FSocket := socket(PF_INET,SOCK_STREAM, 0);
  FIsBlocking := TRUE;
  sockreuse(FSocket,True);
  FListening := false;
  FDisp := nil;
  FCb := nil;
  FOperationObj := TTransportServerOperationImpl.Create;
  FOperationObj.FTranServ := Self;
  FOperation := FOperationObj;
end;

function TTCPTransportServer.accept(): ITransport;
var
  sock : TSocket;
begin
  listen();
  sock := winsock.accept(FSocket,nil,nil);
  result := TTCPTransport.Create();//(sock);
  result.open(sock);
end;

procedure TTCPTransportServer.callback(const disp: IORBDispatcher;event: TDispEvent);
begin
  case event of
    _Read:
       fcb.ts_callback(self,types.tsAccept);
    _Remove: begin
        FDisp := nil;
        FCb := nil;
      end;
    _Moved: FDisp := disp;
    else
      assert(false,'not implemented');
  end;
end;

procedure TTCPTransportServer.listen();
var
  res : integer;
begin
  if not FListening then
    begin
      res := winsock.listen(FSocket,10);
      if res = 0 then FListening := true;
    end;
end;

procedure TTCPTransportServer.block(doblock: boolean);
begin
  if doblock <> FIsBlocking then begin
    FIsBlocking := doblock;
    sockblock(FSocket, doblock);
  end;
end;

procedure TTCPTransportServer.aselect(const disp: IORBDispatcher;
  const cb: ITransportServerCallback);
begin
  if (FDisp <> nil) and (FCb <> nil) then
    begin
      FDisp.remove(self,types._Read);
      FDisp := nil;
      FCb := nil;
    end;
  if cb <> nil then
    begin
      listen();
      disp.socketevent(self,FSocket,types._Read);
      FDisp := disp;
      FCb := cb;
      FOperationObj.FCb := FCb;
    end;
end;

function TTCPTransportServer.bind(const addr: IAddress): Boolean;
var
  sa : TSockAddrIn;
  r : longint;
begin
  sa := addr.socketaddress();
  r := WinSock.bind(FSocket,sa, sizeof (sa));
  result :=  r >= 0;
end;

function TTCPTransportServer.get_address: IAddress;
var
  sa : TSockAddrIn;
  r : longint;
  len: integer;
begin
  result := nil;
  len := SizeOf(sa);
  r := winsock.getsockname(FSocket,sa,len);
  if r < 0 then exit;
  FLocalAddress.set_socketaddress(sa);
  result := FLocalAddress;
end;

procedure seterrno();
var
  e: longint;
begin
  e := WSAGetLastError();
  if e = WSAEINTR then
          errno_var := EINTR
  else if e = WSAEWOULDBLOCK then
          errno_var := EWOULDBLOCK
  else if e = WSAECONNRESET then
          errno_var := ECONNRESET
  else if e = WSAESHUTDOWN then
          errno_var := ECONNRESET
  else
          begin
                  assert(false,'unknown WSAGetLastError() result');
                  errno_var := -1;
          end;
end;

procedure TTCPTransportServer.create_thread;
var
  kt: IWorkerThread;
begin
  // create a new thread
  // there is no limit on how much TransportServer we can have
  // each on runs in its own thread !!
  kt := TPM()._getThreadPool(ord(odAccept)).get_idle_thread();
  kt.registerOperation(Self);
  kt.mark_busy();
end;

procedure TTCPTransportServer.start;
begin
  FOperation.thread.post_state_change();
end;

//*********************************************************************
//  TPrincipal
//*********************************************************************
constructor TPrincipal.Create(transp: ITransport);
begin
  FInfo := Copy(_peer_info);
  FTransport := transp;
end;

class procedure TPrincipal.set_peer_info(info: OctetSeq);
begin
  _peer_info := copy(info);
end;

function TPrincipal.decode(const dec: IDecoder): Boolean;
var
  u : _ulong;
begin
  result := false;
  if not dec.seq_begin(u) then exit;
  setLength(FInfo,u);
  if u > 0 then
    begin
      if dec.buffer.length < u then exit;
      dec.get_octets(Pointer(FInfo)^,u);
    end;
  if not dec.seq_end then exit;
  result := true;
end;

procedure TPrincipal.encode(const enc: IEncoder);
begin
  enc.seq_begin(Length(FInfo));
  if Length(FInfo) > 0 then
    enc.put_octets(Pointer(FInfo)^,Length(FInfo));
  enc.seq_end;
end;

{ TTransportServerOperationImpl }

procedure TTransportServerOperationImpl._run;
begin
  while true do
    FCb.ts_callback(FTranServ, tsAccept);
end;

function TTransportServerOperationImpl.copy: IOperation;
begin

end;

initialization
  errno_var := 0;
end.
