// **********************************************************************
//
// Copyright (c) 2001 MT Tools.
//
// All Rights Reserved
//
// MT_DORB is based in part on the product DORB,
// written by Shadrin Victor
//
// See Readme.txt for contact information
//
// **********************************************************************
unit value;

interface

uses
  value_int,code_int,classes,types;

type

  TValueBase = class(TInterfacedObject,IValueBase)
  protected
    procedure _get_marshal_info(const str: TStrings; out chunked: Boolean); virtual;
    procedure _marshal_members(const enc: IEncoder);  virtual;
    function  _demarshal_members(const dec: Idecoder): Boolean; virtual;
    function narrow_helper(const repoid: PChar): Pointer; virtual;
    class function _create(repoids: TStrings; repoid: string): IValueBase;
  public
    class procedure _marshal(const enc: IEncoder; const vb: IValueBase);
    class function _demarshal(const dec: IDecoder; var vb: IValueBase; const repoid: PChar): Boolean;
  end;


implementation

uses
  orb,orb_int,throw,exceptions;

//***********************************************************************
//  TValueBase
//***********************************************************************
class function TValueBase._create(repoids: TStrings; repoid: string): IValueBase;
var
  dorb: IORB;
  vf : IValueFactory;
  i : integer;
begin
  result := nil;
  dorb := ORB_Instance;
  if repoids.Count = 0 then
    begin
      vf := dorb.lookup_value_factory(PChar(repoid));
      if vf <> nil then
        begin
          result := vf.create_for_umarshal();
          exit;
        end;
    end
  else
    for i := 0 to Pred(repoids.count) do
      begin
        vf := dorb.lookup_value_factory(PChar(repoids[i]));
        if vf <> nil then
        begin
          result := vf.create_for_umarshal();
          exit;
        end;
      end;
  throw.dorb_throw(SystemException.Create_type(st_MARSHAL,0,COMPLETED_MAYBE));
end;

class procedure TValueBase._marshal(const enc: IEncoder; const vb: IValueBase);
var
  lst: TStrings;
  chunked: Boolean;
  vid: longint;
begin
  if vb = nil then
    begin
      enc.value_ref(0);
      exit;
    end;
  vid := enc.valuestate.visited.IndexOf(vb);
  if vid > 0 then
    begin
      enc.value_ref(enc.valuestate.visited[vb]);
      exit;
    end;
  lst := TStringList.Create;
  try
    vb._get_marshal_info(lst,chunked);
    enc.value_begin('',lst,chunked,vid);
    enc.valuestate.visited[vb] := vid;
    vb._marshal_members(enc);
  finally
    lst.Free;
  end;
end;

class function TValueBase._demarshal(const dec: IDecoder; var vb: IValueBase; const repoid: PChar): Boolean;
var
  url: string;
  repoids: TStrings;
  vid : longint;
  is_ref: Boolean;
  temp: IUnknown;
begin
  result := false;
  repoids := TStringList.Create;
  try
    if not dec.value_begin(url,repoids,vid,is_ref) then exit;
    if is_ref then
      begin
        if vid = 0 then
          begin
            vb := nil;
            result := true;
            exit;
          end;
        // indirection
        temp := dec.valuestates.visited[vid];
        if temp = nil then exit;
        vb := IValueBase(temp);
        result := true;
        exit;
      end;
    vb := _create(repoids,repoid);
    if vb = nil then exit;
    dec.valuestates.visited[vid] := vb;
    if not vb._demarshal_members(dec) then exit;
    result := true;
  finally
    repoids.Free;
  end;
end;

procedure TValueBase._get_marshal_info(const str: TStrings; out chunked: Boolean);
begin
  // must be implemented for each valuetype class
  assert(false,'not imlplemented');
end;

procedure TValueBase._marshal_members(const enc: IEncoder);
begin
  // must be implemented for each valuetype class
  assert(false,'not imlplemented');
end;

function  TValueBase._demarshal_members(const dec: Idecoder): Boolean;
begin
  // must be implemented for each valuetype class
  assert(false,'not imlplemented');
  result := false;
end;

function TValueBase.narrow_helper(const repoid: PChar): Pointer;
begin
  result := nil;
end;

end.
